#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Created on Thu Mar 21 18:38:35 2013

@author: Sol
"""
from __future__ import absolute_import, division, print_function

from builtins import str
from past.builtins import basestring
from builtins import object
import os
import inspect
import numbers
from weakref import proxy
import time
import numpy as np
from psychopy import core, misc, colors, logging
import psychopy.tools.colorspacetools as colortools
import psychopy.tools.arraytools as arraytools
import pyglet
pyglet.options['debug_gl'] = False
from pyglet.gl import (glCallList, glFinish, glGenLists, glNewList, glViewport,
                       glMatrixMode, glLoadIdentity, glDisable, glEnable, glColorMaterial,
                       glBlendFunc, glTranslatef, glColor4f, glRectf, glLineWidth, glBegin,
                       GL_LINES, glVertex2d, glEndList, glClearColor, gluOrtho2D, glOrtho,
                       glDeleteLists, GL_COMPILE, GL_PROJECTION, GL_MODELVIEW, glEnd,
                       GL_DEPTH_TEST, GL_BLEND, GL_COLOR_MATERIAL, GL_FRONT_AND_BACK,
                       GL_AMBIENT_AND_DIFFUSE, GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA,
                       glIsEnabled, GL_LINE_SMOOTH, GLint, GLfloat, glGetIntegerv,
                       GL_LINE_WIDTH, glGetFloatv, GL_ALIASED_LINE_WIDTH_RANGE,
                       GL_SMOOTH_LINE_WIDTH_RANGE, GL_SMOOTH_LINE_WIDTH_GRANULARITY,
                       GL_POLYGON_SMOOTH)

from .fontmanager import FontManager
from .textgrid import TextGrid


def getTime():
    return core.getTime()


def is_sequence(arg):
    return (  # not hasattr(arg, "strip") and
        hasattr(arg, "__getitem__") or
        hasattr(arg, "__iter__"))

_system_font_manager = None


def getFontManager(mono_only=True):
    global _system_font_manager
    if _system_font_manager is None:
        _system_font_manager = FontManager(mono_only)
    return _system_font_manager
# check that font manager is working
fm = getFontManager()
font_names = fm.getFontFamilyNames()
if len(font_names) == 0:
    logging.warn("TextBox Font Manager Found No Fonts.")


def getGLInfo():
    gl_info = dict()
    gl_info['GL_LINE_SMOOTH'] = glIsEnabled(GL_LINE_SMOOTH)
    lwidth = GLint()
    glGetIntegerv(GL_LINE_WIDTH, lwidth)
    gl_info['GL_LINE_WIDTH'] = lwidth
    awrange = (GLfloat * 2)(0.0, 0.0)
    glGetFloatv(GL_ALIASED_LINE_WIDTH_RANGE, awrange)
    gl_info['GL_ALIASED_LINE_WIDTH_RANGE'] = awrange[0], awrange[1]
    swrange = (GLfloat * 2)(0.0, 0.0)
    glGetFloatv(GL_SMOOTH_LINE_WIDTH_RANGE, swrange)
    gl_info['GL_SMOOTH_LINE_WIDTH_RANGE'] = swrange[0], swrange[1]
    swg = GLfloat()
    glGetFloatv(GL_SMOOTH_LINE_WIDTH_GRANULARITY, swg)
    gl_info['GL_SMOOTH_LINE_WIDTH_GRANULARITY'] = swg
    return gl_info


class TextBox(object):
    """
    Similar to the visual.TextStim component, TextBox can be used to display
    text within a psychopy window. TextBox and TextStim each have different
    strengths and weaknesses. You should select the most appropriate text
    component type based on how it will be used within the experiment.

    NOTE: As of PsychoPy 1.79, TextBox should be considered experimental.
    The two TextBox demo scripts provided have been tested on
    all PsychoPy supported OS's and run without exceptions. However there are
    very likely bugs in the existing TextBox code and the TextBox API will
    be further enhanced and improved (i.e. changed) over the next couple months.

    TextBox Features:
    ~~~~~~~~~~~~~~~~~

    * Text character placement is very well defined, useful when the exact
      positioning of each letter needs to be known.

    * The text string that is displayed can be changed ( setText() ) and
      drawn ( win.draw() ) **very** quickly. See the TextBox vs. TextStim
      comparison table for details.

    * Built-in font manager; providing easy access to the font family names
      and styles that are available on the computer being used.

    * TextBox is a composite stimulus type, with the following graphical
      elements:
         - TextBox Border / Outline
         - TextBox Fill Area
         - Text Grid Cell Lines
         - Text Glyphs
      Attributes for each of the TextBox graphical elements can be changed
      to control many aspects of how the TextBox is displayed.

    * When using 'rgb' or 'rgb255' color spaces, colors can be specified as
      a list/tuple of 3 elements (red, green, blue), or with four elements
      (reg, green, blue, alpha) which allows different elements of the
      TextBox to use different opacity settings if desired. For colors that
      include the alpha channel value, it will be applied instead of the
      opacity setting of the TextBox, effectively overriding the stimulus
      defined opacity for that part of the textbox graphics. Colors that
      do not include an alpha channel use the opacity setting as normal.

    * Text Line Spacing can be controlled.

    Textbox Limitations:
    ~~~~~~~~~~~~~~~~~~~~

    * Only Monospace Fonts are supported.

    * TextBox component is not a completely **standard** psychopy visual
      stim and has the following functional difference:
          - TextBox attributes are never accessed directly; get* and set*
            methods are always used (this will be changed to use class
            properties in the future).
          - Setting an attribute of a TextBox only supports value replacement,
            ( textbox.setFontColor([1.0,1.0,1.0]) ) and does not support
            specifying operators.

    * Some key word arguments supported by other stimulus types in general,
      or by TextStim itself, are not supported by TextBox. See the TextBox
      class definition for the arguments that are supported.

    * When a new font, style, and size are used it takes about 1 second to
      load and process the font. This is a one time delay for a given
      font name, style, and size. After first being loaded,
      the same font style can be used or re-applied to multiple TextBox
      components with no significant delay.

    * Auto logging or auto drawing is not currently supported.

    TextStim and TextBox Comparison
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    ============================ ============= ===========
    Feature                      TextBox       TextStim
    ============================ ============= ===========
    Change text + redraw time^   1.513 msec    28.537 msec
    No change + redraw time^     0.240 msec    0.931 msec
    Initial Creation time^       0.927 msec    0.194 msec
    MonoSpace Font Support       Yes           Yes
    Non MonoSpace Font Support   No            Yes
    Adjustable Line Spacing      Yes           No
    Precise Text Pos. Info       Yes           No
    Auto logging Support         No            Yes
    Rotation Support             No            Yes
    Word Wrapping Support        Yes           Yes
    ============================ ============= ===========

    ^ Times are in msec.usec format. Tested using the textstim_vs_textbox.py
      demo script provided with the PsychoPy distribution. Results are
      dependent on text length, video card, and OS. Displayed results are
      based on 120 character string with an average of 24 words. Test computer
      used Windows 7 64 bit, PsychoPy 1.79, with a i7 3.4 Ghz CPU, 8 GB RAM,
      and NVIDIA 480 GTX 2GB graphics card.

    Example TextBox Usage:
    ~~~~~~~~~~~~~~~~~~~~~~~

        from psychopy import visual

        win=visual.Window((....)

        # A Textbox stim that will look similar to a TextStim component
        #

        textstimlike=visual.TextBox(
                                    window=win,
                                    text="This textbox looks most like a textstim.",
                                    font_size=18,
                                    font_color=[-1,-1,1],
                                    color_space='rgb',
                                    size=(1.8,.1),
                                    pos=(0.0,.5),
                                    units='norm'
                                   )

        # A Textbox stim that uses more of the supported graphical features
        #
        textboxloaded=visual.TextBox(
                         window=win
                         text='TextBox showing all supported graphical elements',
                         font_size=32,
                         font_color=[1,1,1],
                         border_color=[-1,-1,1], # draw a blue border around stim
                         border_stroke_width=4, # border width of 4 pix.
                         background_color=[-1,-1,-1], # fill the stim background
                         grid_color=[1,-1,-1,0.5], # draw a red line around each
                                                   # possible letter area,
                                                   # 50% transparent
                         grid_stroke_width=1,      # with a width of 1 pix
                         textgrid_shape=[20,2], # specify area of text box
                                                # by the number of cols x
                                                # number of rows of text to support
                                                # instead of by a screen
                                                # units width x height.
                         pos=(0.0,-.5),
                         # If the text string length < num rows * num cols in
                         # textgrid_shape, how should text be justified?
                         #
                         grid_horz_justification='center',
                         grid_vert_justification='center',
                         )

        textstimlike.draw()
        textboxloaded.draw()
        win.flip()

    """
    _textbox_instances = {}
    _gl_info = None

    def __init__(self,
                 window=None,               # PsychoPy Window instance
                 text='Default Test Text.',  # Initial text to be displayed.
                 font_name=None,            # Family name of Font
                 bold=False,                # Bold and italics are used to
                 italic=False,  # determine style of font
                 font_size=32,              # Pt size to use for font.
                 font_color=(0, 0, 0, 1),      # Color to draw the text with.
                 dpi=72,                    # DPI used to create font bitmaps
                 line_spacing=0,            # Amount of extra spacing to add between
                 line_spacing_units='pix',  # lines of text.
                 background_color=None,     # Color to use to fill the entire area
                 # on the screen TextBox is using.
                 border_color=None,         # TextBox border color to use.
                 # Stroke width of TextBox boarder (in pix)
                 border_stroke_width=1,
                 # (width,height) desired for the TextBox
                 size=None,
                 # stim to use. Specify using the unit
                 # type the textBox is using.
                 # (cols,rows) of characters to use when
                 textgrid_shape=None,
                 # creating the textgrid layout.
                 # rows*cols = maximum number of chars
                 # that can be displayed. If textgrid_shape
                 # is not None, then the TextBox size
                 # must be at least large enough to hold
                 # the number of specified cols and rows.
                 # If the size specified is less than
                 # what is needed, the size will be increased
                 # automatically.
                 # (x,y) screen position for the TextBox
                 pos=(0.0, 0.0),
                 # stim. Specify using the unit
                 # type the textBox is using.
                 align_horz='center',       # Determines how TextBox x pos is
                 # should be interpreted to.
                 # 'left', 'center', 'right' are valid options.
                 align_vert='center',       # Determines how TextBox y pos is
                 # should be interpreted to.
                 # 'left', 'center', 'right' are valid options.
                 units='norm',                # Coordinate unit type to use for position
                 # and size related attributes. Valid
                 # options are 'pix', 'cm', 'deg', 'norm'
                 # Only pix is currently working though.
                 grid_color=None,           # Color to draw the TextBox text grid
                 # lines with.
                 # Line thickness (in pix) to use when
                 grid_stroke_width=1,
                 # displaying text grid lines.
                 color_space='rgb',          # PsychoPy color space to use for any
                 # color attributes of TextBox.
                 opacity=1.0,               # Opacity (transparency) to use for
                 # TextBox graphics, assuming alpha
                 # channel was not specified in the color
                 # attribute.
                 grid_horz_justification='left',  # 'left', 'center', 'right'
                 grid_vert_justification='top',  # 'top', 'bottom', 'center'
                 autoLog=True,              # Log each time stim is updated.
                 interpolate=False,
                 name=None
                 ):
        self._window = proxy(window)

        self._font_name = font_name
        self._font_size = font_size
        self._dpi = dpi
        self._bold = bold
        self._italic = italic

        self._text = text
        self._label = name
        self._line_spacing = line_spacing
        self._line_spacing_units = line_spacing_units
        self._border_color = border_color
        self._background_color = background_color
        self._border_stroke_width = border_stroke_width
        self._grid_horz_justification = grid_horz_justification
        self._grid_vert_justification = grid_vert_justification
        self._align_horz = align_horz
        self._align_vert = align_vert
        self._size = size
        self._position = pos
        self._textgrid_shape = textgrid_shape
        self._interpolate = interpolate

        self._draw_start_dlist = None
        self._draw_end_dlist = None
        self._draw_te_background_dlist = None

        if TextBox._gl_info is None:
            TextBox._gl_info = getGLInfo()

        aliased_wrange = TextBox._gl_info['GL_ALIASED_LINE_WIDTH_RANGE']
        antia_wrange = TextBox._gl_info['GL_SMOOTH_LINE_WIDTH_RANGE']
        if grid_stroke_width and grid_color:
            if interpolate:
                if grid_stroke_width < antia_wrange[0]:
                    self._grid_stroke_width = antia_wrange[0]
                if grid_stroke_width > antia_wrange[1]:
                    self._grid_stroke_width = antia_wrange[1]
            else:
                if grid_stroke_width < aliased_wrange[0]:
                    self._grid_stroke_width = aliased_wrange[0]
                if grid_stroke_width > aliased_wrange[1]:
                    self._grid_stroke_width = aliased_wrange[1]
        if border_stroke_width and border_color:
            if interpolate:
                if border_stroke_width < antia_wrange[0]:
                    self._border_stroke_width = antia_wrange[0]
                if border_stroke_width > antia_wrange[1]:
                    self._border_stroke_width = antia_wrange[1]
            else:
                if border_stroke_width < aliased_wrange[0]:
                    self._border_stroke_width = aliased_wrange[0]
                if border_stroke_width > aliased_wrange[1]:
                    self._border_stroke_width = aliased_wrange[1]

        self._units = units
        if self._units is None:
            self._units = self._window.units

        self._opacity = opacity
        if opacity is None:
            self._opacity = 1.0
        elif float(opacity) and float(opacity) >= 0 and float(opacity) <= 1.0:
            self._opacity = float(opacity)
        else:
            raise ValueError(
                "Text Box: opacity must be a number between 0.0 and 1.0, or None (which == 1.0). %s is not valid" % (str(opacity)))

        self._color_space = color_space
        if self._color_space is None:
            self._color_space = self._window.colorSpace

        # TODO: Implement support for autoLog
        self._auto_log = autoLog

        self._current_glfont = None
        self._text_grid = None

        if self._label is None:
            self._label = 'TextBox_%s' % (str(int(time.time())))

        fm = getFontManager()
        if self._font_name is None:
            self._font_name = fm.getFontFamilyStyles()[0][0]
        gl_font = fm.getGLFont(
            self._font_name, self._font_size, self._bold, self._italic, self._dpi)
        self._current_glfont = gl_font

        self._text_grid = TextGrid(self, line_color=grid_color,
                                   line_width=grid_stroke_width, font_color=list(
                                       font_color),
                                   shape=textgrid_shape,
                                   grid_horz_justification=grid_horz_justification,
                                   grid_vert_justification=grid_vert_justification)
        self._text_grid.setCurrentFontDisplayLists(
            gl_font.charcode2displaylist)

        self._text = self._text.replace('\r\n', '\n')
        self._text_grid._createParsedTextDocument(self._text)

        self._textbox_instances[self.getLabel()] = proxy(self)

    def getWindow(self):
        """
        Returns the psychopy window that the textBox is associated with.
        """
        return self._window

    def getText(self):
        """
        Return the text to display.
        """
        return self._text

    def setText(self, text_source):
        """
        Set the text to be displayed within the Textbox.

        Note that once a TextBox has been created, the number of character
        rows and columns is static. To change the size of a TextBox,
        a new TextBox stim must be created to replace the current Textbox stim.
        Therefore ensure that the textbox is large enough to display
        the largest length string to be presented in the TextBox. Characters
        that do not fit within the TextBox will not be displayed.

        Color value must be valid for the color space being used by the TextBox.
        """
        if not self._text:
            raise ValueError(
                "TextBox.setText only accepts strings with a length > 0")

        self._text = text_source.replace('\r\n', '\n')
        return self._text_grid._setText(self._text)

    def getDisplayedText(self):
        """
        Return the text that fits within the TextBox and therefore is actually
        seen. This is equal to:

            text_length=len(self.getText())
            cols,rows=self.getTextGridShape()

            displayed_text=self.getText()[0:min(text_length,rows*cols]
        """
        return self._getTextWrappedDoc().getDisplayedText()

    def getTextGridCellPlacement(self):
        """
        Returns a 3d numpy array containing position information for each text grid cell
        in the TextBox. The array has the shape (num_cols,num_rows,cell_bounds),
        where num_cols is the number of textgrid columns in the TextBox.
        num_rows is the number of textgrid rows in the TextBox. cell_bounds is
        a 4 element array containing the (x pos, y pos, width, height) data
        for the given cell. Position fields are for the top left hand corner of
        the cell box. Column and Row indices start at 0.

        To get the shape of the textgrid in terms of columns and rows, use:

        cell_pos_array=textbox.getTextGridCellPlacement()
        col_row_count=cell_pos_array.shape[:2]

        To access the position, width, and height for textgrid cell at
        column 0 and row 0 (so the top left cell in the textgrid):

        cell00=cell_pos_array[0,0,:]

        For the cell at col 3, row 1 (so 4th cell on second row):

        cell41=cell_pos_array[4,1,:]
        """
        col_lines = self._text_grid._col_lines
        row_lines = self._text_grid._row_lines

        cellinfo = np.zeros(
            (len(col_lines[:-1]), len(row_lines[:-1]), 4), dtype=np.float32)

        tb_tl = self._getTopLeftPixPos()
        tg_tl = self._text_grid._position
        # print('tb_tl,tg_tl:',tb_tl,tg_tl)
        starting_x, starting_y = tb_tl[0] + tg_tl[0], tb_tl[1] - tg_tl[1]
        #print('Text Cell Placement:')

        for i, x in enumerate(col_lines[:-1]):
            for j, y in enumerate(row_lines[:-1]):
                px, py = starting_x + x, starting_y + y
                w, h = col_lines[i + 1] - px, -(row_lines[j + 1] - py)
                cellinfo[i, j, 0] = px
                cellinfo[i, j, 1] = py
                cellinfo[i, j, 2] = w
                cellinfo[i, j, 3] = h

        for i, x in enumerate(col_lines[:-1]):
            for j, y in enumerate(row_lines[:-1]):
                x, y = self._pix2units(
                    (float(cellinfo[i, j, 0]), float(cellinfo[i, j, 1])))
                w, h = self._pix2units(
                    (float(cellinfo[i, j, 2]), float(cellinfo[i, j, 3])), False)
                cellinfo[i, j, 0] = x
                cellinfo[i, j, 1] = y
                cellinfo[i, j, 2] = w
                cellinfo[i, j, 3] = h
                # print(col %d, row %d: '%(i,j),cellinfo[i,j,:])

#        print('cell[0,0]:',cellinfo[0,0,:])
        return cellinfo

    def getTextGridCellForCharIndex(self, char_index):
        return self._getTextWrappedDoc().getTextGridCellForCharIndex(char_index)

    def getGlyphPositionForTextIndex(self, char_index):
        """
        For the provided char_index, which is the index of one character in
        the current text being displayed by the TextBox ( getDisplayedText() ),
        return the bounding box position, width, and height for the associated
        glyph drawn to the screen. This factors in the glyphs position within
        the textgrid cell it is being drawn in, so the returned bounding box is
        for the actual glyph itself, not the textgrid cell. For textgrid cell placement
        information, see the getTextGridCellPlacement() method.

        The glyph position for the given text index is returned as a tuple
        (x,y,width,height), where x,y is the top left hand corner of the bounding box.

        Special Cases:
            * If the index provided is out of bounds for the currently displayed text, None is returned.
            * For u' ' (space) characters, the full textgrid cell bounding box is returned.
            * For u'\n' ( new line ) characters,the textgrid cell bounding box is returned, but with the box width set to 0.
        """

        if char_index < 0 or char_index >= len(self.getDisplayedText()):
            raise IndexError(
                "The provided index of %d is out of range for the currently displayed text." % (char_index))

        # Get the Glyph info for the char in question:
        gl_font = getFontManager().getGLFont(self._font_name, self._font_size,
                                             self._bold, self._italic, self._dpi)
        glyph_data = gl_font.charcode2glyph.get(ord(self._text[char_index]))
        ox, oy = glyph_data['offset'][
            0], gl_font.max_ascender - glyph_data['offset'][1]
        gw, gh = glyph_data['size']
#        print('glyph_data for %s: %d,%d %d,%d'%(glyph_data['unichar'],ox,oy,gw,gh))

        # get the col,row for the xhar index provided
        col_row = self._getTextWrappedDoc().getTextGridCellForCharIndex(char_index)
        if col_row is None:
            return None
        cline = self._getTextWrappedDoc().getParsedLine(col_row[1])
        # print('cline._trans_left,cline._trans_top:',cline._trans_left,cline._trans_top)
        c = col_row[0] + cline._trans_left
        r = col_row[1] + cline._trans_top
        x, y, width, height = self.getTextGridCellPlacement()[c, r, :]
#        print('text_grid cell (%d,%d) placement: %d,%d %d,%d'%(c,r,x,y,width,height))
#        print('gygh bounds: ',x+ox,y-oy,gw,gh)
        ox, oy = self._pix2units((ox, oy), False)
        gw, gh = self._pix2units((gw, gh), False)
        #print('glyph_data %d %d -> %.3f %.3f'%(glyph_data['size'][0],glyph_data['size'][1],gw,gh))
        return x + ox, y - oy, gw, gh

    def _getTextWrappedDoc(self):
        return self._text_grid._text_document

    def getPosition(self):
        """
        Return the x,y position of the textbox, in getUnitType() coord space.
        """
        return self._position

    def setPosition(self, pos):
        """
        Set the (x,y) position of the TextBox on the Monitor. The position must
        be given using the unit coord type used by the stim.

        The TextBox position is interpreted differently depending on the
        Horizontal and Vertical Alignment settings of the stim. See
        getHorzAlignment() and getVertAlignment() for more information.


        For example,
        if the TextBox alignment is specified as left, top, then the position
        specifies the top left hand corner of where the stim will be drawn.
        An alignment of bottom,right indicates that the position value will
        define where the bottom right corner of the TextBox will be drawn.
        A horz., vert. alignment of center, center will place the center of
        the TextBox at pos.
        """
        if pos[0] != self._position[0] or pos[1] != self._position[1]:
            self._position = pos[0], pos[1]
            self._deleteBackgroundDL()
            self._deleteStartDL()

    def getUnitType(self):
        """
        Returns which of the psychopy coordinate systems are used by the
        TextBox. Position and size related attributes mush be specified
        relative to the unit type being used. Valid options are:
            - pix
            - norm
            - cm
        """
        return self._units

    def getHorzAlign(self):
        """
        Return what textbox x position should be interpreted as. Valid options
        are 'left', 'center', or 'right' .
        """
        return self._align_horz

    def setHorzAlign(self, v):
        """
        Specify how the horizontal (x) component of the TextBox position
        is to be interpreted. left = x position is the left edge, right =
        x position is the right edge x position, and center = the x position
        is used to center the stim horizontally.
        """
        if v != self._align_horz:
            self._align_horz = v
            self._deleteBackgroundDL()
            self._deleteStartDL()

    def getVertAlign(self):
        """
        Return what textbox y position should be interpreted as. Valid options
        are 'top', 'center', or 'bottom' .
        """
        return self._align_vert

    def setVertAlign(self, v):
        """
        Specify how the vertical (y) component of the TextBox position
        is to be interpreted. top = y position is the top edge, bottom =
        y position is the bottom edge y position, and center = the y position
        is used to center the stim vertically.
        """
        if v != self._align_vert:
            self._align_vert = v
            self._deleteBackgroundDL()
            self._deleteStartDL()

    def getSize(self):
        """
        Return the width,height of the TextBox, using the unit type being
        used by the stimulus.
        """
        return self._size

    def getFontColor(self):
        """
        Return the color used when drawing text glyphs.
        """
        return self._text_grid._font_color

    def setFontColor(self, c):
        """
        Set the color to use when drawing text glyphs within the TextBox.
        Color value must be valid for the color space being used by the TextBox.
        For 'rgb', 'rgb255', and 'norm' based colors, three or four element lists
        are valid. Three element colors use the TextBox getOpacity() value to
        determine the alpha channel for the color. Four element colors use the
        value of the fourth element to set the alpha value for the color.
        """
        if c != self._text_grid._font_color:
            self._text_grid._font_color = c
            self._text_grid._deleteTextDL()

    def getBorderColor(self):
        """
        A border can be drawn around the perimeter of the TextBox. This method
        sets the color of that border.
        """
        return self._border_color

    def setBorderColor(self, c):
        """
        Set the color to use for the border of the textBox. The TextBox border
        is a rectangular outline drawn around the edges of the TextBox stim.
        Color value must be valid for the color space being used by the TextBox.

        A value of None will disable drawing of the border.
        """
        if c != self._border_color:
            self._border_color = c
            self._deleteBackgroundDL()

    def getBackgroundColor(self):
        """
        Get the color used to fill the rectangular area of the TextBox stim.
        All other graphical elements of the TextBox are drawn on top of the
        background.
        """
        return self._background_color

    def setBackgroundColor(self, c):
        """
        Set the fill color used to fill the rectangular area of the TextBox stim.
        Color value must be valid for the color space being used by the TextBox.

        A value of None will disable drawing of the TextBox background.
        """
        if c != self._background_color:
            self._background_color = c
            self._deleteBackgroundDL()

    def getTextGridLineColor(self):
        """
        Return the color used when drawing the outline of the text grid cells.
        Each letter displayed in a TextBox populates one of the text cells
        defined by the shape of the TextBox text grid.
        Color value must be valid for the color space being used by the TextBox.

        A value of None indicates drawing of the textgrid lines is disabled.
        """
        return self._text_grid._line_color

    def setTextGridLineColor(self, c):
        """
        Set the color used when drawing text grid lines.
        These are lines that can be drawn which mark the bounding box for
        each character within the TextBox text grid.
        Color value must be valid for the color space being used by the TextBox.

        Provide a value of None to disable drawing of textgrid lines.
        """
        if c != self._text_grid._line_color:
            self._text_grid._line_color = c
            self._text_grid._deleteGridLinesDL()

    def getColorSpace(self):
        """
        Returns the psychopy color space used when specifying colors for
        the TextBox. Supported values are:
            - 'rgb'
            - 'rbg255'
            - 'norm'
            - hex (implicit)
            - html name (implicit)
        See the Color Space section of the PsychoPy docs for details.
        """
        return self._color_space

    def getHorzJust(self):
        """
        Return how text should laid out horizontally when the
        number of columns of each text grid row is greater than the number
        needed to display the text for that text row.
        """
        return self._text_grid._horz_justification

    def setHorzJust(self, v):
        """
        Specify how text within the TextBox should be aligned horizontally.
        For example, if a text grid has 10 columns, and the text being displayed
        is 6 characters in length, the horizontal justification determines
        if the text should be draw starting at the left of the text columns (left),
        or should be centered on the columns ('center', in this example
        there would be two empty text cells to the left and right of the text.),
        or should be drawn such that the last letter of text is drawn in the
        last column of the text row ('right').
        """
        if v != self._text_grid._horz_justification:
            self._text_grid.setHorzJust(v)
            self._text_grid._deleteTextDL()

    def getVertJust(self):
        """
        Return how text should laid out vertically when the
        number of text grid rows is greater than the number
        needed to display the current text
        """
        return self._text_grid._vert_justification

    def setVertJust(self, v):
        """
        Specify how text within the TextBox should be aligned vertically.
        For example, if a text grid has 3 rows for text, and the text being
        displayed all fits on one row, the vertical justification determines
        if the text should be draw on the top row of the text grid (top),
        or should be centered on the rows ('center', in this example
        there would be one row  above and below the row used to draw the text),
        or should be drawn on the last row of the text grid, ('bottom').
        """
        if v != self._text_grid._vert_justification:
            self._text_grid.setVertJust(v)
            self._text_grid._deleteTextDL()

    def getBorderWidth(self):
        """
        Get the stroke width of the optional TextBox area outline. This is always
        given in pixel units.
        """
        return self._border_stroke_width

    def setBorderWidth(self, c):
        """
        Set the stroke width (in pixels) to use for the border of the TextBox
        stim. Border values must be within the range of stroke widths supported
        by the OpenGL driver used by the graphics. Setting the
        width outside the valid range will result in the stroke width being
        clamped to the nearest end of the valid range.

        Use the TextBox.getValidStrokeWidths() to access the minimum -
        maximum range of valid line widths.
        """
        if c != self._border_stroke_width:
            if self._interpolate:
                lrange = TextBox._gl_info['GL_SMOOTH_LINE_WIDTH_RANGE']
            else:
                lrange = TextBox._gl_info['GL_ALIASED_LINE_WIDTH_RANGE']

            if c < lrange[0]:
                c = lrange[0]
            elif c > lrange[1]:
                c = lrange[1]

            self._border_stroke_width = c
            self._deleteBackgroundDL()

    def getTextGridLineWidth(self):
        """
        Return the stroke width (in pixels) of the optional lines drawn around
        the text grid cell areas.
        """
        return self._text_grid._line_width

    def setTextGridLineWidth(self, c):
        """
        Set the stroke width (in pixels) to use for the text grid character
        bounding boxes. Border values must be within the range of stroke
        widths supported by the OpenGL driver used by the computer graphics
        card. Setting the width outside the valid range will result in the
        stroke width being clamped to the nearest end of the valid range.

        Use the TextBox.getGLineRanges() to access a dict containing some
        OpenGL parameters which provide the minimum, maximum, and resolution
        of valid line widths.
        """
        if c != self._text_grid._line_width:
            if self._interpolate:
                lrange = TextBox._gl_info['GL_SMOOTH_LINE_WIDTH_RANGE']
            else:
                lrange = TextBox._gl_info['GL_ALIASED_LINE_WIDTH_RANGE']

            if c < lrange[0]:
                c = lrange[0]
            elif c > lrange[1]:
                c = lrange[1]

            self._text_grid._line_width = c
            self._text_grid._deleteGridLinesDL()

    def getValidStrokeWidths(self):
        """
        Returns the stroke width range supported by the graphics card being
        used. If the TextBox is Interpolated, a tuple is returns using
        float values, with the following structure:

        ((min_line_width, max_line_width), line_width_granularity)

        If Interpolation is disabled for the TextBox, the returned tuple elements
        are int values, with the following structure:

        (min_line_width, max_line_width)

        """
        if self._interpolate:
            return (TextBox._gl_info['GL_SMOOTH_LINE_WIDTH_RANGE'],
                    TextBox._gl_info['GL_SMOOTH_LINE_WIDTH_GRANULARITY'])
        else:
            return self._gl_info['GL_ALIASED_LINE_WIDTH_RANGE']

    def getOpacity(self):
        """
        Get the default TextBox transparency level used for color related
        attributes. 0.0 equals fully transparent, 1.0 equals fully opaque.
        """
        return self._opacity

    def setOpacity(self, o):
        """
        Sets the TextBox transparency level to use for color related
        attributes of the Textbox. 0.0 equals fully transparent, 1.0 equals
        fully opaque.

        If opacity is set to None, it is assumed to have a default value of 1.0.

        When a color is defined with a 4th element in the colors element list,
        then this opacity value is ignored and the alpha value provided in the
        color itself is used for that TextGrid element instead.
        """
        if o != self._opacity and o >= 0.0 and o <= 1.0:
            self._text_grid._deleteTextDL()
            self._deleteBackgroundDL()
            self._text_grid._deleteGridLinesDL()
            self._deleteStartDL()
            self._deleteEndDL()
            self._opacity = o

    def getInterpolated(self):
        """
        Returns whether interpolation is enabled for the TextBox
        when it is drawn. When True, GL_LINE_SMOOTH and GL_POLYGON_SMOOTH
        are enabled within OpenGL; otherwise they are disabled.
        """
        return self._interpolate

    def setInterpolated(self, interpolate):
        """
        Specify whether interpolation should be enabled for the TextBox
        when it is drawn. When interpolate == True, GL_LINE_SMOOTH and
        GL_POLYGON_SMOOTH are enabled within OpenGL. When interpolate is set
        to False, GL_POLYGON_SMOOTH and GL_LINE_SMOOTH are disabled.
        """
        if interpolate != self._interpolate:
            self._deleteStartDL()
            self._interpolate = interpolate

    def getLabel(self):
        """
        Return the label / name assigned to the textbox. This does not impact
        how the stimulus looks when drawn, and instead is used for internal
        purposes only.
        """
        return self._label

    def getName(self):
        """
        Same as the GetLabel method.
        """
        return self._label

    def getAutoLog(self):
        """
        Indicates if changes to textBox attribute values should be logged
        automatically by PsychoPy. *Currently not supported by TextBox.
        """
        return self._auto_log

    def setAutoLog(self, v):
        """
        Specify if changes to textBox attribute values should be logged
        automatically by PsychoPy. True enables auto logging; False disables it.
        *Currently not supported by TextBox.
        """
        self._auto_log = v

    def getLineSpacing(self):
        """
        Return the additional spacing being applied between rows of text.
        The value is in units specified by the textbox getUnits() method.
        """
        return self._line_spacing

    def draw(self):
        """
        Draws the TextBox to the back buffer of the graphics card. Then call
        win.flip() to display the changes drawn. If draw() is not called prior
        to a call to win.flip(), the textBox will not be displayed for that
        retrace.
        """
        self._te_start_gl()
        self._te_bakground_dlist()
        self._text_grid._text_glyphs_gl()
        self._text_grid._textgrid_lines_gl()
        self._te_end_gl()

    def _te_start_gl(self):
        if not self._draw_start_dlist:
            dl_index = glGenLists(1)
            glNewList(dl_index, GL_COMPILE)
            glViewport(0, 0, self._window.size[0],
                       self._window.size[1])
            glMatrixMode(GL_PROJECTION)
            glLoadIdentity()
            glOrtho(0, self._window.size[0], 0,
                    self._window.size[1], -1, 1)
            glMatrixMode(GL_MODELVIEW)
            glLoadIdentity()
            self._window.depthTest = False
            glEnable(GL_BLEND)
            glEnable(GL_COLOR_MATERIAL)
            glColorMaterial(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE)
            glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA)
            if self._interpolate:
                glEnable(GL_LINE_SMOOTH)
                glEnable(GL_POLYGON_SMOOTH)
            else:
                glDisable(GL_LINE_SMOOTH)
                glDisable(GL_POLYGON_SMOOTH)
            t, l = self._getTopLeftPixPos()
            glTranslatef(t, l, 0)
            glEndList()
            self._draw_start_dlist = dl_index
        glCallList(self._draw_start_dlist)

    def _deleteStartDL(self):
        if self._draw_start_dlist:
            glDeleteLists(self._draw_start_dlist, 1)
            self._draw_start_dlist = None

    def _te_bakground_dlist(self):
        if not self._draw_te_background_dlist and (self._background_color or self._border_color):
            dl_index = glGenLists(1)
            glNewList(dl_index, GL_COMPILE)

            # draw textbox_background and outline
            # t,l=self._getTopLeftPixPos()
            #glTranslatef(t,l, 0 )
            border_thickness = self._border_stroke_width
            size = self._getPixelSize()
            if self._border_stroke_width is None:
                border_thickness = 0
            if self._background_color:
                bcolor = self._toRGBA(self._background_color)
                glColor4f(*bcolor)
                # size=self._getPixelSize()
                glRectf(0, 0, size[0], -size[1])
            if self._border_color:
                glLineWidth(border_thickness)
                bcolor = self._toRGBA(self._border_color)
                glColor4f(*bcolor)
                glBegin(GL_LINES)
                x1 = 0
                y1 = 0
                x2, y2 = size
                x2, y2 = x2, y2
                hbthick = border_thickness // 2
                if hbthick < 1:
                    hbthick = 1
                glVertex2d(x1 - border_thickness, y1 + hbthick)
                glVertex2d(x2 + border_thickness, y1 + hbthick)
                glVertex2d(x2 + hbthick, y1)
                glVertex2d(x2 + hbthick, -y2)
                glVertex2d(x2 + border_thickness, -y2 - hbthick)
                glVertex2d(x1 - border_thickness, -y2 - hbthick)
                glVertex2d(x1 - hbthick, -y2)
                glVertex2d(x1 - hbthick, y1)
                glEnd()
            glColor4f(0.0, 0.0, 0.0, 1.0)
            glEndList()
            self._draw_te_background_dlist = dl_index
        if (self._background_color or self._border_color):
            glCallList(self._draw_te_background_dlist)

    def _deleteBackgroundDL(self):
        if self._draw_te_background_dlist:
            glDeleteLists(self._draw_te_background_dlist, 1)
            self._draw_te_background_dlist = None

    def _te_end_gl(self):
        if not self._draw_end_dlist:
            dl_index = glGenLists(1)
            glNewList(dl_index, GL_COMPILE)

            rgb = self._window.rgb
            rgb = TextBox._toRGBA2(
                rgb, 1, self._window.colorSpace, self._window)
            glClearColor(rgb[0], rgb[1], rgb[2], 1.0)
            glViewport(0, 0, int(self._window.size[0]), int(
                self._window.size[1]))
            glMatrixMode(GL_PROJECTION)  # Reset The Projection Matrix
            glLoadIdentity()
            gluOrtho2D(-1, 1, -1, 1)
            glMatrixMode(GL_MODELVIEW)  # Reset The Projection Matrix
            glLoadIdentity()

            glEndList()
            self._draw_end_dlist = dl_index
        glCallList(self._draw_end_dlist)

    def _deleteEndDL(self):
        if self._draw_end_dlist:
            glDeleteLists(self._draw_end_dlist, 1)
            self._draw_end_dlist = None

    @staticmethod
    def _toPix(xy, units, window):
        if isinstance(xy, numbers.Number):
            xy = xy, xy
        elif is_sequence(xy):
            if len(xy) == 1:
                xy = xy[0], xy[0]
            else:
                xy = xy[:2]
        else:
            return ValueError("TextBox: coord variables must be array-like or a single number. Invalid: %s" % (str(xy)))

        if not isinstance(xy[0], numbers.Number) or not isinstance(xy[1], numbers.Number):
            return ValueError("TextBox: coord variables must only contain numbers. Invalid: %s" % (str(xy)))

        if units in ('pix', 'pixs'):
            return xy
        if units in ['deg', 'degs']:
            return misc.deg2pix(xy[0], window.monitor), misc.deg2pix(xy[1], window.monitor)
        if units in ['cm']:
            return misc.cm2pix(xy[0], window.monitor), misc.cm2pix(xy[1], window.monitor)
        if units in ['norm']:
            # -1.0 to 1.0
            if xy[0] <= 1.0 and xy[0] >= -1.0 and xy[1] <= 1.0 and xy[1] >= -1.0:
                return xy[0] * window.size[0] / 2.0, xy[1] * window.size[1] / 2.0

        return ValueError("TextBox: %s, %s could not be converted to pix units" % (str(xy), str(units)))

    def _pix2units(self, xy, is_position=True):
        units = self._units
        window = self._window
        ww, wh = float(window.size[0]), float(window.size[1])

        if isinstance(xy, numbers.Number):
            xy = xy, xy
        elif is_sequence(xy):
            if len(xy) == 1:
                xy = xy[0], xy[0]
            else:
                xy = xy[:2]
        else:
            raise ValueError(
                "TextBox: coord variables must be array-like or a single number. Invalid: %s" % (str(xy)))

        if not isinstance(xy[0], numbers.Number) or not isinstance(xy[1], numbers.Number):
            raise ValueError("TextBox: coord variables must only contain numbers. Invalid: %s, %s, %s" % (
                str(xy), str(type(xy[0])), str(type(xy[1]))))

        x, y = xy
        if is_position:
            # convert to psychopy pix, origin is center of monitor.
            # print('x,y:',x,y)
            x, y = int(x - ww / 2), int(y - wh / 2)
            #print('xy psycho:',x,y)
            # print('---')
        if units in ('pix', 'pixs'):
            return x, y
        if units in ['deg', 'degs']:
            return misc.pix2deg(x, window.monitor), misc.deg2pix(y, window.monitor)
        if units in ['cm']:
            return misc.pix2cm(x, window.monitor), misc.cm2pix(y, window.monitor)
        if units in ['norm']:
            return x / ww * 2.0, y / wh * 2.0

        raise ValueError(
            "TextBox: %s, %s could not be converted to pix units" % (str(xy), str(units)))

    def _toRGBA(self, color):
        return self.__class__._toRGBA2(color, self._opacity, self._color_space, self._window)

    @classmethod
    def _toRGBA2(cls, color, opacity=None, color_space=None, window=None):

        if color is None:
            raise ValueError("TextBox: None is not a valid color input")
        if not colors.isValidColor(color):
            raise ValueError(
                "TextBox: %s is not a valid color." % (str(color)))

        valid_opacity = opacity >= 0.0 and opacity <= 1.0
        if isinstance(color, basestring):
            if color[0] == '#' or color[0:2].lower() == '0x':
                rgb255color = colors.hex2rgb255(color)
                if rgb255color and valid_opacity:
                    return rgb255color[0] / 255.0, rgb255color[1] / 255.0, rgb255color[2] / 255.0, opacity
                else:
                    raise ValueError(
                        "TextBox: %s is not a valid hex color." % (str(color)))

            named_color = colors.colors.get(color.lower())
            if named_color and valid_opacity:
                return (named_color[0] + 1.0) / 2.0, (named_color[1] + 1.0) / 2.0, (named_color[2] + 1.0) / 2.0, opacity
            raise ValueError(
                "TextBox: String color value could not be translated: %s" % (str(color)))

        if isinstance(color, (float, int, int)) or (is_sequence(color) and len(color) == 3):
            color = arraytools.val2array(color, length=3)
            if color_space == 'dkl' and valid_opacity:
                dkl_rgb = None
                if window:
                    dkl_rgb = window.dkl_rgb
                rgb = colortools.dkl2rgb(color, dkl_rgb)
                return (rgb[0] + 1.0) / 2.0, (rgb[1] + 1.0) / 2.0, (rgb[2] + 1.0) / 2.0, opacity
            if color_space == 'lms' and valid_opacity:
                lms_rgb = None
                if window:
                    lms_rgb = window.lms_rgb
                rgb = colortools.lms2rgb(color, lms_rgb)
                return (rgb[0] + 1.0) / 2.0, (rgb[1] + 1.0) / 2.0, (rgb[2] + 1.0) / 2.0, opacity
            if color_space == 'hsv' and valid_opacity:
                rgb = colortools.hsv2rgb(color)
                return (rgb[0] + 1.0) / 2.0, (rgb[1] + 1.0) / 2.0, (rgb[2] + 1.0) / 2.0, opacity
            if color_space == 'rgb255' and valid_opacity:
                rgb = color
                if [cc for cc in color if cc < 0 or cc > 255]:
                    raise ValueError(
                        'TextBox: rgb255 colors must contain elements between 0 and 255. Value: ' + str(rgb))
                return rgb[0] / 255.0, rgb[1] / 255.0, rgb[2] / 255.0, opacity
            if color_space == 'rgb' and valid_opacity:
                rgb = color
                if [cc for cc in color if cc < -1.0 or cc > 1.0]:
                    raise ValueError(
                        'TextBox: rgb colors must contain elements between -1.0 and 1.0. Value: ' + str(rgb))
                return (rgb[0] + 1.0) / 2.0, (rgb[1] + 1.0) / 2.0, (rgb[2] + 1.0) / 2.0, opacity

        if is_sequence(color) and len(color) == 4:
            if color_space == 'rgb255':
                if [cc for cc in color if cc < 0 or cc > 255]:
                    raise ValueError(
                        'TextBox: rgb255 colors must contain elements between 0 and 255. Value: ' + str(color))
                return color[0] / 255.0, color[1] / 255.0, color[2] / 255.0, color[3] / 255.0
            if color_space == 'rgb':
                if [cc for cc in color if cc < -1.0 or cc > 1.0]:
                    raise ValueError(
                        'TextBox: rgb colors must contain elements between -1.0 and 1.0. Value: ' + str(color))
                return (color[0] + 1.0) / 2.0, (color[1] + 1.0) / 2.0, (color[2] + 1.0) / 2.0, (color[3] + 1.0) / 2.0

        raise ValueError("TextBox: color: %s, opacity: %s, is not a valid color for color space %s." % (
            str(color), str(opacity), color_space))

    def _reset(self):
        self._text_grid.reset()

    def _getPixelSize(self):
        if self._units == 'norm':
            r = self._toPix(
                (self._size[0] - 1.0, self._size[1] - 1.0), self._units, self._window)
            return int(r[0] + self._window.size[0] / 2), int(r[1] + self._window.size[1] / 2)
        return [int(x) for x in self._toPix(self._size, self._units, self._window)]

    def _setSize(self, pix_sz):
        units = self._units
        if units in ('pix', 'pixs'):
            self._size = list(pix_sz)
        if units in ['deg', 'degs']:
            self._size = misc.pix2deg(pix_sz[0], self._window.monitor), misc.pix2deg(
                pix_sz[1], self._window.monitor)
        if units in ['cm']:
            self._size = misc.pix2cm(pix_sz[0], self._window.monitor), misc.pix2cm(
                pix_sz[1], self._window.monitor)
        if units in ['norm']:
            pw, ph = pix_sz
            dw, dh = self._window.size
            nw = (pw / float(dw)) * 2.0
            nh = (ph / float(dh)) * 2.0
            self._size = nw, nh

    def _getPixelPosition(self):
        ppos = self._toPix(self._position, self._units, self._window)
        return int(ppos[0]), int(ppos[1])

    def _getPixelTextLineSpacing(self):
        if self._line_spacing:
            max_size = self._current_glfont.max_tile_width, self._current_glfont.max_tile_height
            line_spacing_units = self._line_spacing_units
            line_spacing_height = self._line_spacing

            if line_spacing_units == 'ratio':
                # run though _toPix to validate line_spacing value type only
                r = self._toPix(line_spacing_height, 'pix', self._window)[0]
                return int(max_size[1] * r)

            return self._toPix(line_spacing_height, line_spacing_units, self._window)
        return 0

    def _getTopLeftPixPos(self):
        # Create a window position based on the window size, alignment types,
        #   TextBox size, etc...
        win_w, win_h = self._window.size
        te_w, te_h = self._getPixelSize()
        te_x, te_y = self._getPixelPosition()
        # convert te_x,te_y from psychopy pix coord to gl pix coord
        te_x, te_y = te_x + win_w // 2, te_y + win_h // 2
        # convert from alignment to top left
        horz_align, vert_align = self._align_horz, self._align_vert
        if horz_align.lower() == u'center':
            te_x = te_x - te_w // 2
        elif horz_align.lower() == u'right':
            te_x = te_x - te_w
        if vert_align.lower() == u'center':
            te_y = te_y + te_h // 2
        if vert_align.lower() == u'bottom':
            te_y = te_y + te_h
        return te_x, te_y

    def __del__(self):
        if hasattr(self, '_textbox_instance') and self.getName() in self._textbox_instance:
            del self._textbox_instances[self.getName()]
        del self._current_glfont
        del self._text_grid



# -*- coding: utf-8 -*-
# ioHub DataStore to Pandas DataFrames - Event Filtering Support
# Part of the psychopy.iohub library.
# .. moduleauthor:: Sol Simpson <sol@isolver-software.com> and
#                   Pierce Edmiston <pierce.edmiston@gmail.com>
# Copyright (C) 2012-2016 iSolver Software Solutions
# Distributed under the terms of the GNU General Public License (GPL).
from __future__ import division, absolute_import, print_function

import numpy as np
import pandas as pd


class InterestPeriodDefinition(object):
    """InterestPeriodDefinition Class.

    Parent class of InterestPeriodDefinition implementations.

    """
    next_ipid = 1

    def __init__(self, name=None):
        self._name = name
        self._ipid = self.next_ipid
        self.__class__.next_ipid += 1
        if name is None:
            self._name = self.__class__.__name__ + '_' + str(self._ipid)

        self._ip_df = None

    @property
    def name(self):
        return self._name

    @property
    def ipid(self):
        return self._ipid

    def find(self, target, ip_cols=None):
        df = target[:]
        df['ip_id_num'] = np.nan
        df['ip_id'] = self.ipid
        df['ip_name'] = self.name
        df = df.groupby(level=[0, 1], group_keys=False).apply(self.__find)

        if ip_cols is not None:
            df = self._merge_ip_cols(df, ip_cols)

        return df

    def __find(self, target_group):
        group_id = target_group.index[0]
        group_ips = self.ip_df.ix[group_id]
        if len(group_ips) > 0:
            overlapping = []
            for index, ip in group_ips.iterrows():
                in_ip = ((target_group['time'] >= ip['start_time']) &
                         (target_group['time'] <= ip['end_time']))
                target_in_ip = target_group[in_ip]
                target_in_ip['ip_id_num'] = ip['ip_id_num']
                overlapping += [target_in_ip]
            target_filtered = pd.concat(overlapping)
        else:
            target_filtered = target_group[[False] * len(target_group)]
        return target_filtered

    def filter(self, target, ip_cols=None):
        df = target[:]
        df['ip_id'] = self.ipid
        df['ip_name'] = self.name
        df = target.groupby(
            level=[
                0, 1], group_keys=False).apply(
            self.__filter)

        if ip_cols is not None:
            df = self._merge_ip_cols(df, ip_cols)

        return df

    def __filter(self, group):
        """
        HT http://stackoverflow.com/a/21370058/2506078
        """
        ips = self.ip_df.ix[group.index[0]]
        start_idx = np.searchsorted(ips['start_time'], group['time']) - 1
        end_idx = np.searchsorted(ips['end_time'], group['time'])
        mask = (start_idx == end_idx)
        group['ip_id_num'] = np.nan
        group['ip_id_num'][mask] = start_idx
        return group[mask]

    def _merge_ip_cols(self, target, cols):
        if not isinstance(cols, dict):
            if not hasattr(cols, '__iter__'):
                cols = [cols]
            cols = dict(zip(cols, cols))

        temp_target = target.set_index('ip_id_num', append=True)
        temp_ips = self.ip_df.set_index('ip_id_num', append=True)
        temp_ips = temp_ips[cols.keys()]

        temp_target = temp_target.merge(
            temp_ips, left_index=True, right_index=True)
        temp_target = temp_target.rename(columns=cols)
        temp_target = temp_target.reset_index('ip_id_num')
        return temp_target

    def _extract_criteria_match(self, source, criteria, return_cols, exact):
        col = list(criteria.keys())[0]  # eventually we'll want to allow for
        val = criteria[col]      # multiple criteria matches

        if exact:
            is_match = (source[col] == val)
        else:
            is_match = (source[col].str.contains(val))

        if not isinstance(return_cols, dict):
            return_cols = dict(zip(return_cols, return_cols))

        keep_cols = list(return_cols.keys())
        matches = source[is_match][keep_cols].rename(columns=return_cols)

        return matches

    def __enumerate_ips(self, group):
        group['ip_id_num'] = range(len(group))
        return group

    def _ip_zipper(self, start, end, temp_index='ip_id_num'):
        # TODO: make sure the two dfs "zip" nicely
        _start = start.groupby(level=[0, 1]).apply(self.__enumerate_ips)
        _end = end.groupby(level=[0, 1]).apply(self.__enumerate_ips)

        _start.set_index(temp_index, append=True, inplace=True)
        _end.set_index(temp_index, append=True, inplace=True)

        _all = pd.merge(_start, _end, left_index=True, right_index=True)
        return _all.reset_index(temp_index)

#############################################


class EventBasedIP(InterestPeriodDefinition):
    """EventBasedIP Class.

    trial_ip=BoundingEventsIP(name='trial_ip',
                              start_source_df=exp_data.MESSAGE,
                              start_criteria={'text':'TRIAL_START'},
                              end_source_df=exp_data.MESSAGE,
                              end_criteria={'text':'TRIAL_END'}
                              )

    """
    next_ipid = 1

    def __init__(
            self,
            name=None,
            start_source_df=None,
            start_criteria=None,
            end_source_df=None,
            end_criteria=None,
            exact=True):
        InterestPeriodDefinition.__init__(self, name)

        self._start_source_df = start_source_df
        self._end_source_df = end_source_df or start_source_df[:]
        self._start_criteria = start_criteria
        self._end_criteria = end_criteria
        self._exact = exact

    @property
    def ip_df(self):
        """Return a Pandas DF where each row represents a start and end time
        instance that matches the InterestPeriodDefinition criteria.

        ip_df is indexed on:
            experiment_id
            session_id

        ip_df has columns:
            start_time
            start_event_id
            end_time
            end_event_id
            ip_name
            ip_id
            ip_id_num

        """
        if self._ip_df is None:
            # Match start_source_df[start_criteria.key]==start_criteria.value
            # Currently only support ip criteria in the form of
            # {*_df_col_name:*_df_col_value} for ip start time and ip end time
            # selections, where rows from start_source_df will be selected when:
            #
            #   start_source_df[df_col_name] == df_col_value,
            #
            # and rows from end_source_df will be selected when:
            #
            #   start_source_df[start_df_col_name] == start_df_col_value

            start_cols = {'time': 'start_time', 'event_id': 'start_event_id'}
            start_ip_df = self._extract_criteria_match(
                self.start_source_df,
                criteria=self.start_criteria,
                return_cols=start_cols,
                exact=self._exact)

            end_cols = {'time': 'end_time', 'event_id': 'end_event_id'}
            end_ip_df = self._extract_criteria_match(
                self.end_source_df,
                criteria=self.end_criteria,
                return_cols=end_cols,
                exact=self._exact)

            self._ip_df = self._ip_zipper(start_ip_df, end_ip_df)

            # Add ip identifier cols
            self._ip_df['ip_name'] = self.name
            self._ip_df['ip_id'] = self.ipid

        return self._ip_df

    @property
    def start_source_df(self):
        return self._start_source_df

    @start_source_df.setter
    def start_source_df(self, v):
        self._start_source_df = v
        self._ip_df = None

    @property
    def end_source_df(self):
        return self._end_source_df

    @end_source_df.setter
    def end_source_df(self, v):
        self._end_source_df = v
        self._ip_df = None

    @property
    def start_criteria(self):
        return self._start_criteria

    @start_criteria.setter
    def start_criteria(self, v):
        self._start_criteria = v
        self._ip_df = None

    @property
    def end_criteria(self):
        return self._end_criteria

    @end_criteria.setter
    def end_criteria(self, v):
        self._end_criteria = v
        self._ip_df = None


##########################################

class MessageBasedIP(EventBasedIP):
    next_ipid = 1

    def __init__(self, name=None, message_df=None, start_text='TRIAL_START',
                 end_text='TRIAL_END', exact=True):
        EventBasedIP.__init__(self, name, message_df,
                              start_criteria={'text': start_text},
                              end_criteria={'text': end_text},
                              exact=exact)

##########################################


class ConditionVariableBasedIP(InterestPeriodDefinition):
    """ConditionVariableBasedIP Class.

    cv_ip=ConditionVariableBasedIP(name='cv_ip',
                        source_df=[some df],
                      start_col_name='TRIAL_START',
                      end_col_name='TRIAL_END'
                      criteria={}
                      )

    """
    next_ipid = 1

    def __init__(
            self,
            name=None,
            source_df=None,
            start_col_name=None,
            end_col_name=None,
            criteria=[]):
        InterestPeriodDefinition.__init__(self, name)

        self._source_df = source_df
        self._start_col_name = start_col_name
        self._end_col_name = end_col_name
        self._criteria = criteria

    @property
    def ip_df(self):
        """Return a Pandas DF where each row represents a start and end time
        instance that matches the InterestPeriodDefinition criteria.

        ip_df is indexed on:
            experiment_id
            session_id

        ip_df has columns:
            start_time
            end_time
            ip_name
            ip_id
            ip_id_num

        """
        if self._ip_df is None:
            self._ip_df = self._source_df[
                [self._start_col_name, self.end_col_name]]
            self._ip_df = self._ip_df.rename(
                columns={
                    self._start_col_name: 'start_time',
                    self.end_col_name: 'end_time'})

            for a_critera in self._criteria:
                # TODO: Support filtering of cond_var rows based on criteria
                pass

            # Add ip identifier cols
            self._ip_df['ip_name'] = self.name
            self._ip_df['ip_id'] = self.ipid
            self._ip_df['ip_id_num'] = range(1, len(self._ip_df) + 1)

        return self._ip_df

    @property
    def source_df(self):
        return self._source_df

    @property
    def start_col_name(self):
        return self._start_col_name

    @start_col_name.setter
    def start_col_name(self, v):
        self._start_col_name = v
        self._ip_df = None

    @property
    def end_col_name(self):
        return self._end_col_name

    @end_col_name.setter
    def end_col_name(self, v):
        self._end_col_name = v
        self._ip_df = None

    @property
    def criteria(self):
        return self._criteria

    @criteria.setter
    def criteria(self, v):
        self._criteria = v
        self._ip_df = None



"""
Demo using a custom serial rx parser to generate ioHub Serial Device events.

The parseserial.py file is also required for this demo, as it contains the
custom parser function that the ioHub Serial device uses during runtime.

** This demo assumes that whatever is written out to the serial port is what the
serial device receives back as rx data. **

"""
from __future__ import print_function
import time
from psychopy import core, visual
from psychopy.iohub import launchHubServer

# Settings for serial port communication.
SERIAL_PORT = 'COM226'
BAUDRATE = 19200

# event_parser_info dict:
#
# parser_function key value can be a str giving the module.function path,
# or it can be the actual function object to be run by the iohub process.
#
# *Important:* The function provided should be in a file that can be imported
# as a module without causing unwanted behavior on the iohub process.
# Some options:
#     1) Put the function in a file that contains only the function,
#        as is done in this example.
#     2) Ensure any script logic that will be run when the file is called by
#        a user ( i.e. python.exe filewithfunc.py ) is inside a:
#            if __name__ == '__main__':
#        condition so it is not run when the file is only imported.

event_parser_info = dict(parser_function="parseserial.checkForSerialEvents",
                         parser_kwargs=dict(var1='not used', var2=1234))
# configure iohub
exp_code = 'serial_demo'
sess_code = 'S_{0}'.format(int(time.mktime(time.localtime())))
iohubkwargs = {'experiment_code': exp_code,
               'session_code': sess_code,
               'serial.Serial': dict(name='serial',
                                     port=SERIAL_PORT,
                                     baud=BAUDRATE,
                                     parity='NONE',
                                     bytesize=8,
                                     event_parser=event_parser_info)}

# start the iohub server and set up display and PST box devices
io = launchHubServer(**iohubkwargs)
serial_device = io.devices.serial
keyboard = io.devices.keyboard

# Start collecting data from the PST box in the background.
serial_device.enableEventReporting(True)

# Create a window.
win = visual.Window((1024, 768), units='pix')

# Instruction text.
instruction = visual.TextStim(win, text='Monitoring for serial input events....\n\nPress any key to exit.')
# Display instruction.
instruction.draw()
win.flip()
io.clearEvents('all')

# Check for keyboard and serial events.
# Exit on keyboard press event.
# Print any serial events.
#
while not keyboard.getPresses():
    serial_device.write('aaa')
    io.wait(0.05)
    serial_device.write('bbb')
    io.wait(0.05)
    serial_device.write('ccc')
    io.wait(.500)
    for serevt in serial_device.getEvents():
        print(serevt)

# Stop recording events from the PST box and switch off all lamps.
serial_device.enableEventReporting(False)

# Close the window and quit the program.
io.quit()
core.quit()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Coder is the python IDE for PsychoPy
"""

from __future__ import absolute_import, print_function
from .coder import CoderFrame, BaseCodeEditor  # pylint: disable=W0401



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Demo for the class psychopy.visual.Pie().

Use the `Pie` class to create a Kanizsa figure which produces illusory
contours.

"""
from psychopy import core
import psychopy.visual as visual
from psychopy.visual import Pie
from psychopy import event

# open a window to render the shape
win = visual.Window((600, 600), allowGUI=False, monitor='testMonitor')

# create the stimulus object
pieStim = Pie(
    win, radius=50, start=0., end=270., fillColor=(-1., -1., -1.), units='pix')

message = visual.TextStim(
    win, text='Any key to quit', pos=(0, -0.8), units='norm')

# positions of the corners of the shapes
pos = [(-100, 100), (100, 100), (-100, -100), (100, -100)]

# orientations of the shapes
ori = [180., 270., 90., 0.]

while not event.getKeys():

    for i in range(4):
        pieStim.pos = pos[i]
        pieStim.ori = ori[i]
        pieStim.draw()

    message.draw()
    win.flip()

win.close()
core.quit()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Part of the PsychoPy library
# Copyright (C) 2002-2018 Jonathan Peirce (C) 2019-2020 Open Science Tools Ltd.
# Distributed under the terms of the GNU General Public License (GPL).

from .functions import logInPavlovia, logOutPavlovia
from psychopy.localization import _translate
from psychopy.projects import pavlovia
from psychopy import prefs
import os
import wx
try:
    import wx.lib.agw.hyperlink as wxhl  # 4.0+
except ImportError:
    import wx.lib.hyperlink as wxhl # <3.0.2


class UserEditor(wx.Dialog):
    defStyle = wx.DEFAULT_DIALOG_STYLE #| wx.RESIZE_BORDER
    def __init__(self, parent=None, id=wx.ID_ANY, style=defStyle,
                 *args, **kwargs):

        wx.Dialog.__init__(self, parent, id,
                           *args, style=style, **kwargs)
        self.SetWindowStyle(wx.STAY_ON_TOP)  # this will be a modal dialog
        panel = wx.Panel(self, wx.ID_ANY, style=wx.TAB_TRAVERSAL)
        self.parent = parent
        self.app = wx.GetApp()
        pavSession = pavlovia.getCurrentSession()
        if pavSession.user:
            pavSession.gitlab.auth()
            self.user = pavSession.user
        else:
            self.user = logInPavlovia(parent=parent)
            if not self.user:
                return  # they were given a login but cancelled
        if type(self.user) != pavlovia.User:
            self.user = pavlovia.User(gitlabData=self.user)

        # create the controls
        userField = wxhl.HyperLinkCtrl(panel, id=wx.ID_ANY,
                                       label=self.user.url, URL=self.user.url)
        logoutBtn = wx.Button(panel, label="Logout")
        logoutBtn.Bind(wx.EVT_BUTTON, self.onLogout)
        nameLabel = wx.StaticText(panel, id=wx.ID_ANY,
                                  label=_translate("Full name:"))
        self.nameField = wx.StaticText(panel, wx.ID_ANY, self.user.name)
        bitmapFile = self.user.avatar or "user_invisible.png"
        self.avatarBtn = self.app.iconCache.makeBitmapButton(
                    parent=panel,
                    filename=bitmapFile,
                    label=self.user.name, name=self.user.name)

        org = self.user.organization or ""
        orgLabel = wx.StaticText(panel, wx.ID_ANY, _translate("Organization:"))
        self.orgField = wx.StaticText(panel, wx.ID_ANY, org, size=(300, -1))

        bio = self.user.bio or ""
        self.bioLabel = wx.StaticText(panel, wx.ID_ANY, _translate("Bio (250 chrs):"))
        self.bioField = wx.StaticText(panel, wx.ID_ANY, bio,
                                    style=wx.TE_MULTILINE)
        # submit/cancel
        buttonMsg = _translate("OK")
        updateBtn = wx.Button(panel, id=wx.ID_OK, label=buttonMsg)
        updateBtn.Bind(wx.EVT_BUTTON, self.submitChanges)
        # cancelBtn = wx.Button(panel, id=wx.ID_CANCEL, label=_translate("Cancel"))
        # cancelBtn.Bind(wx.EVT_BUTTON, self.onCancel)

        # layout
        userAndLogout = wx.BoxSizer(wx.VERTICAL)
        userAndLogout.Add(userField, 1, wx.ALL | wx.CENTER, 5)
        userAndLogout.Add(logoutBtn, 0, wx.ALL | wx.CENTER , 5)
        topRow = wx.BoxSizer(wx.HORIZONTAL)
        topRow.Add(userAndLogout, 1, wx.ALL | wx.CENTER, 5)
        topRow.Add(self.avatarBtn, 0, wx.ALL | wx.RIGHT, 5)

        fieldsSizer = wx.FlexGridSizer(cols=2, rows=5, vgap=10, hgap=10)
        fieldsSizer.AddMany([
            (nameLabel, 0, wx.ALIGN_RIGHT | wx.ALIGN_CENTER_VERTICAL),
            (self.nameField,0, wx.EXPAND),
            (orgLabel, 0, wx.ALIGN_RIGHT | wx.ALIGN_CENTER_VERTICAL),
            (self.orgField,0, wx.EXPAND),
            (self.bioLabel, 0, wx.ALIGN_RIGHT | wx.ALIGN_CENTER_VERTICAL),
            (self.bioField, 1, wx.EXPAND),
        ])

        btnSizer = wx.BoxSizer(wx.HORIZONTAL)
        btnSizer.Add(updateBtn)

        border = wx.BoxSizer(wx.VERTICAL)
        border.Add(topRow, 0, wx.ALL| wx.EXPAND, 5)
        border.Add(fieldsSizer, 1, wx.ALL | wx.EXPAND, 10)
        border.Add(btnSizer, 0, wx.ALIGN_RIGHT | wx.ALL, 5)
        panel.SetSizerAndFit(border)
        self.Fit()

    def onLogout(self, evt=None):
        logOutPavlovia(self.parent)
        self.user = None
        self.Close()

    def onCancel(self, evt=None):
        self.EndModal(wx.ID_CANCEL)

    def submitChanges(self, evt=None):
        # print("updating gitlab.pavlovia.org users from within PsychoPy not yet working")
        # isDirty = False
        # toCheck = {'name': self.nameField.GetValue,
        #            'organization': self.orgField.GetValue,
        #            'bio': self.bioField.GetValue,
        #            }
        # for field in toCheck:
        #     newVal = toCheck[field]()
        #     prev = getattr(self.user, field)
        #     if prev != newVal and not (prev is None and newVal==""):
        #         setattr(self.user, field, newVal)
        #         isDirty = True
        # if isDirty:
        #     # currently gives an error:
        #     # https://github.com/python-gitlab/python-gitlab/issues/547
        #     self.user.save()
        self.EndModal(wx.ID_OK)

    def onSetAvatar(self, event=None):
        print("Uploading a user image (avatar) from within PsychoPy is not yet supported. "
              "You can do that by going to the gitlab.pavlovia.org profile settings.")

    def onURL(self, evt):
        print(dir(evt))



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Demo for using labjack DAC devices

See also
    http: //labjack.com/support/labjackpython
but note that the version shipped with standalone PsychoPy
has u3 (and others below an umbrella called labjack) so the import
line is slightly different to the documentation on LabJack's website
"""

from __future__ import absolute_import, division, print_function

from builtins import range
from psychopy import visual, core, event, sound
try:
    from labjack import u3
except ImportError:
    import u3

# sound.setAudioAPI('pyaudio')

win = visual.Window([800, 800])
stim = visual.GratingStim(win, color=-1, sf=0)
snd = sound.Sound(880)
print(snd)
# setup labjack U3
ports = u3.U3()
FIO4 = 6004  # the address of line FIO4

while True:
    # do this repeatedly for timing tests
    ports.writeRegister(FIO4, 0)  # start low

    # draw black square
    stim.draw()
    win.flip()

    # wait for a key press
    if 'q' in event.waitKeys():
        break

    # set to white, flip window and raise level port FIO4
    stim.setColor(1)
    stim.draw()
    win.flip()
    ports.writeRegister(FIO4, 1)
    snd.play()
    for frameN in range(4):
        stim.draw()
        win.flip()

    # set color back to black and set FIO4 to low again
    stim.setColor(-1)
    stim.draw()
    win.flip()
    ports.writeRegister(FIO4, 0)

win.close()
core.quit()

# The contents of this file are in the public domain.



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
parallel ports demo

This is for win32 only.
"""

from __future__ import absolute_import, division, print_function

from builtins import range
from psychopy import visual, core
from psychopy import parallel

nFramesOn = 5
nFramesOff = 30
nCycles = 2
parallel.setPortAddress(0x378)  # address for parallel port on many machines
pinNumber = 2  # choose a pin to write to (2-9).

# setup the stimuli and other objects we need
win = visual.Window([1280, 1024], allowGUI=False)  # make a window
win.flip()  # present it
myStim = visual.GratingStim(win, tex=None, mask=None, color='white', size=2)
myClock = core.Clock()  # just to keep track of time

# present a stimulus for EXACTLY 20 frames and exactly 5 cycles
for cycleN in range(nCycles):
    for frameN in range(nFramesOff):
        # don't draw, just refresh the window
        win.flip()
        parallel.setData(0)  # sets all pins low

    for frameN in range(nFramesOn):
        myStim.draw()
        win.flip()
        # immediately *after* screen refresh set pins as desired
        parallel.setPin(2, 1)  # sets just this pin to be high

# report the mean time afterwards
print('total time=%0.6f' % myClock.getTime())
print('avg frame rate=%0.3f' % win.fps())

# set pins back to low
win.flip()
parallel.setData(0)  # sets all pins low again

win.close()
core.quit()

# The contents of this file are in the public domain.



"""ioHub Common Eye Tracker Interface for EyeLink(C) Systems"""
# Part of the psychopy.iohub library.
# Copyright (C) 2012-2016 iSolver Software Solutions
# Distributed under the terms of the GNU General Public License (GPL).
from .eyetracker import (EyeTracker, MonocularEyeSampleEvent,
                        BinocularEyeSampleEvent, FixationStartEvent,
                        FixationEndEvent, SaccadeStartEvent,
                        SaccadeEndEvent, BlinkStartEvent,BlinkEndEvent)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Take a "screen-shot" (full or partial), save to a ImageStim()-like
RBGA object.`"""

# Part of the PsychoPy library
# Copyright (C) 2002-2018 Jonathan Peirce (C) 2019-2020 Open Science Tools Ltd.
# Distributed under the terms of the GNU General Public License (GPL).

from __future__ import absolute_import, division, print_function

from builtins import str

# Ensure setting pyglet.options['debug_gl'] to False is done prior to any
# other calls to pyglet or pyglet submodules, otherwise it may not get picked
# up by the pyglet GL engine and have no effect.
# Shaders will work but require OpenGL2.0 drivers AND PyOpenGL3.0+
import pyglet
pyglet.options['debug_gl'] = False
GL = pyglet.gl

import psychopy  # so we can get the __path__
from psychopy import core, logging

# tools must only be imported *after* event or MovieStim breaks on win32
# (JWP has no idea why!)
from psychopy.tools.attributetools import attributeSetter, setAttribute
from psychopy.tools.typetools import float_uint8
from psychopy.visual.image import ImageStim

try:
    from PIL import Image
except ImportError:
    from . import Image

import numpy


class BufferImageStim(ImageStim):
    """Take a "screen-shot", save as an ImageStim (RBGA object).

    The screen-shot is a single collage image composed of static elements
    that you can treat as being a single stimulus. The screen-shot can be of
    the visible screen (front buffer) or hidden (back buffer).

    BufferImageStim aims to provide fast rendering, while still allowing
    dynamic orientation, position, and opacity. It's fast to draw but
    slower to init (same as an ImageStim).

    You specify the part of the screen to capture (in norm units), and
    optionally the stimuli themselves (as a list of items to be drawn).
    You get a screenshot of those pixels. If your OpenGL does not support
    arbitrary sizes, the image will be larger, using square powers of two
    if needed, with the excess image being invisible (using alpha). The
    aim is to preserve the buffer contents as rendered.

    Checks for OpenGL 2.1+, or uses square-power-of-2 images.

    **Example**::

        # define lots of stimuli, make a list:
        mySimpleImageStim = ...
        myTextStim = ...
        stimList = [mySimpleImageStim, myTextStim]

        # draw stim list items & capture (slow; see EXP log for times):
        screenshot = visual.BufferImageStim(myWin, stim=stimList)

        # render to screen (very fast, except for the first draw):
        while <conditions>:
            screenshot.draw()  # fast; can vary .ori, .pos, .opacity
            other_stuff.draw() # dynamic
            myWin.flip()

    See coder Demos > stimuli > bufferImageStim.py for a demo, with timing stats.

    :Author:
        - 2010 Jeremy Gray, with on-going fixes
    """

    def __init__(self, win, buffer='back', rect=(-1, 1, 1, -1),
                 sqPower2=False, stim=(), interpolate=True,
                 flipHoriz=False, flipVert=False, mask='None', pos=(0, 0),
                 name=None, autoLog=None):
        """
        :Parameters:

            buffer :
                the screen buffer to capture from, default is 'back' (hidden).
                'front' is the buffer in view after win.flip()
            rect :
                a list of edges [left, top, right, bottom] defining a
                screen rectangle which is the area to capture from the
                screen, given in norm units.
                default is fullscreen: [-1, 1, 1, -1]
            stim :
                a list of item(s) to be drawn to the back buffer (in order).
                The back buffer is first cleared (without the win being
                flip()ed), then stim items are drawn, and finally the buffer
                (or part of it) is captured. Each item needs to have its
                own .draw() method, and have the same window as win.
            interpolate :
                whether to use interpolation (default = True, generally good,
                especially if you change the orientation)
            sqPower2 :
                - False (default) = use rect for size if OpenGL = 2.1+
                - True = use square, power-of-two image sizes
            flipHoriz :
                horizontally flip (mirror) the captured image, default = False
            flipVert :
                vertically flip (mirror) the captured image; default = False
        """
        # depends on: window._getRegionOfFrame

        # what local vars are defined (these are the init params) for use by
        # __repr__
        self._initParams = dir()
        self._initParams.remove('self')

        self.autoLog = False  # set this False first and change later
        _clock = core.Clock()
        if stim:  # draw all stim to the back buffer
            win.clearBuffer()
            buffer = 'back'
            if hasattr(stim, '__iter__'):
                for stimulus in stim:
                    try:
                        if stimulus.win == win:
                            stimulus.draw()
                        else:
                            msg = ('BufferImageStim.__init__: user '
                                   'requested "%s" drawn in another window')
                            logging.warning(msg % repr(stimulus))
                    except AttributeError:
                        msg = 'BufferImageStim.__init__: "%s" failed to draw'
                        logging.warning(msg % repr(stimulus))
            else:
                raise ValueError('Stim is not iterable in BufferImageStim. '
                                 'It should be a list of stimuli.')

        # take a screenshot of the buffer using win._getRegionOfFrame():
        glversion = pyglet.gl.gl_info.get_version()
        if glversion >= '2.1' and not sqPower2:
            region = win._getRegionOfFrame(buffer=buffer, rect=rect)
        else:
            if not sqPower2:
                msg = ('BufferImageStim.__init__: defaulting to square '
                       'power-of-2 sized image (%s)')
                logging.debug(msg % glversion)
            region = win._getRegionOfFrame(buffer=buffer, rect=rect,
                                           squarePower2=True)
        if stim:
            win.clearBuffer()

        # turn the RGBA region into an ImageStim() object:
        if win.units in ['norm']:
            pos *= win.size / 2.

        size = region.size / win.size / 2.
        super(BufferImageStim, self).__init__(
            win, image=region, units='pix', mask=mask, pos=pos,
            size=size, interpolate=interpolate, name=name, autoLog=False)
        self.size = region.size

        # to improve drawing speed, move these out of draw:
        self.desiredRGB = self._getDesiredRGB(
            self.rgb, self.colorSpace, self.contrast)
        self.thisScale = numpy.array([4, 4])
        self.flipHoriz = flipHoriz
        self.flipVert = flipVert

        # set autoLog now that params have been initialised
        wantLog = autoLog is None and self.win.autoLog
        self.__dict__['autoLog'] = autoLog or wantLog
        if self.autoLog:
            logging.exp("Created %s = %s" % (self.name, str(self)))
            msg = 'BufferImageStim %s: took %.1fms to initialize'
            logging.exp(msg % (name, 1000 * _clock.getTime()))

    @attributeSetter
    def flipHoriz(self, flipHoriz):
        """If set to True then the image will be flipped horizontally
        (left-to-right). Note that this is relative to the original image,
        not relative to the current state.
        """
        self.__dict__['flipHoriz'] = flipHoriz

    @attributeSetter
    def flipVert(self, flipVert):
        """If set to True then the image will be flipped vertically
        (left-to-right). Note that this is relative to the original image,
        not relative to the current state.
        """
        self.__dict__['flipVert'] = flipVert

    def setFlipHoriz(self, newVal=True, log=None):
        """Usually you can use 'stim.attribute = value' syntax instead,
        but use this method if you need to suppress the log message.
        """
        setAttribute(self, 'flipHoriz', newVal, log)  # call attributeSetter

    def setFlipVert(self, newVal=True, log=None):
        """Usually you can use 'stim.attribute = value' syntax instead,
        but use this method if you need to suppress the log message.
        """
        setAttribute(self, 'flipVert', newVal, log)  # call attributeSetter

    def draw(self, win=None):
        """Draws the BufferImage on the screen, similar to
        :class:`~psychopy.visual.ImageStim` `.draw()`.
        Allows dynamic position, size, rotation, mirroring, and opacity.
        Limitations / bugs: not sure what happens with shaders and
        self._updateList()
        """
        if win is None:
            win = self.win
        self._selectWindow(win)

        GL.glPushMatrix()  # preserve state
        # GL.glLoadIdentity()

        # dynamic flip
        GL.glScalef(self.thisScale[0] * (1, -1)[self.flipHoriz],
                    self.thisScale[1] * (1, -1)[self.flipVert], 1.0)

        # enable dynamic position, orientation, opacity; depth not working?
        GL.glColor4f(self.desiredRGB[0], self.desiredRGB[1],
                     self.desiredRGB[2], self.opacity)

        GL.glCallList(self._listID)  # make it happen
        GL.glPopMatrix()  # return the view to previous state



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Part of the PsychoPy library
# Copyright (C) 2002-2018 Jonathan Peirce (C) 2019-2020 Open Science Tools Ltd.
# Distributed under the terms of the GNU General Public License (GPL).


class DependencyError(Exception):
    """The user requested something that won't be possible because
    of a dependency error (e.g. audiolib that isn't available)
    """
    pass


class SoundFormatError(Exception):
    """The user tried to create two streams (diff sample rates) on a machine
    that won't allow that
    """
    pass


class NoUserError(Exception):
    pass


class ConnectionError(Exception):
    pass


class NoGitError(DependencyError):
    pass


class ConditionsImportError(Exception):
    pass

class MissingFontError(Exception):
    pass



# -*- coding: utf-8 -*-
"""
Shows how to create two textBox stim and present them. The first textbox
simply displays the text provided, centered. The second textbox shows more
of the configuration options available for the stim type.

Created on Thu Mar 21 18:37:10 2013

@author: Sol
"""
from psychopy import visual, core, event

# Create Window
window=visual.Window((800,600),
                        units='norm',
                        fullscr=False, allowGUI=True,
                        screen=0
                        )

sometext=u'PRESS ANY KEY TO QUIT DEMO.'                                                        
textbox=visual.TextBox(window=window, 
                         text=sometext,
                         font_size=21,
                         font_color=[-1,-1,1], 
                         size=(1.9,.3),
                         pos=(0.0,0.25), 
                         grid_horz_justification='center',
                         units='norm',
                         )

textbox2=visual.TextBox(window=window,
                         text='This TextBox illustrates many of the different UX elements.', 
                         font_size=32,
                         font_color=[1,-1,-1], 
                         background_color=[-1,-1,-1,1],
                         border_color=[-1,-1,1,1],
                         border_stroke_width=4,
                         textgrid_shape=[20,4], # 20 cols (20 chars wide)
                                                # by 4 rows (4 lines of text)
                         pos=(0.0,-0.25),
                         )

textbox.draw()
textbox2.draw()
demo_start=window.flip()     

event.clearEvents()
last_attrib_change_time=demo_start
while True:
    if core.getTime()-last_attrib_change_time> 2.5:
        last_attrib_change_time=core.getTime()

    textbox.draw()
    textbox2.draw()

    # Update the display to show any stim changes
    flip_time=window.flip()

    # End the test when a keyboard event is detected
    #
    kb_events=event.getKeys()
    if kb_events:
        break

core.quit()


#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Part of the PsychoPy library
# Copyright (C) 2002-2018 Jonathan Peirce (C) 2019-2020 Open Science Tools Ltd.
# Distributed under the terms of the GNU General Public License (GPL).

"""shaders programs for either pyglet or pygame
"""

from __future__ import absolute_import, print_function

import pyglet.gl as GL
import psychopy.tools.gltools as gltools
from ctypes import c_int, c_char_p, c_char, cast, POINTER, byref


class Shader:
    def __init__(self, vertexSource=None, fragmentSource=None):

        def compileShader(source, shaderType):
            """Compile shader source of given type (only needed by compileProgram)
            """
            shader = GL.glCreateShaderObjectARB(shaderType)
            # if Py3 then we need to convert our (unicode) str into bytes for C
            if type(source) != bytes:
                source = source.encode()
            prog = c_char_p(source)
            length = c_int(-1)
            GL.glShaderSourceARB(shader,
                                 1,
                                 cast(byref(prog), POINTER(POINTER(c_char))),
                                 byref(length))
            GL.glCompileShaderARB(shader)

            # check for errors
            status = c_int()
            GL.glGetShaderiv(shader, GL.GL_COMPILE_STATUS, byref(status))
            if not status.value:
                print_log(shader)
                GL.glDeleteShader(shader)
                raise ValueError('Shader compilation failed')
            return shader

        self.handle = GL.glCreateProgramObjectARB()

        if vertexSource:
            vertexShader = compileShader(
                vertexSource, GL.GL_VERTEX_SHADER_ARB
            )
            GL.glAttachObjectARB(self.handle, vertexShader)
        if fragmentSource:
            fragmentShader = compileShader(
                fragmentSource, GL.GL_FRAGMENT_SHADER_ARB
            )
            GL.glAttachObjectARB(self.handle, fragmentShader)

        GL.glValidateProgramARB(self.handle)
        GL.glLinkProgramARB(self.handle)

        if vertexShader:
            GL.glDeleteObjectARB(vertexShader)
        if fragmentShader:
            GL.glDeleteObjectARB(fragmentShader)

    def bind(self):
        GL.glUseProgram(self.handle)

    def unbind(self):
        GL.glUseProgram(0)

    def setFloat(self, name, value):
        if type(name) is not bytes:
            name = bytes(name, 'utf-8')
        loc = GL.glGetUniformLocation(self.handle, name)
        if not hasattr(value, '__len__'):
            GL.glUniform1f(loc, value)
        elif len(value) in range(1, 5):
            # Select the correct function
            { 1 : GL.glUniform1f,
              2 : GL.glUniform2f,
              3 : GL.glUniform3f,
              4 : GL.glUniform4f
              # Retrieve uniform location, and set it
            }[len(value)](loc, *value)
        else:
            raise ValueError("Shader.setInt '{}' should be length 1-4 not {}"
                             .format(name, len(value)))

    def setInt(self, name, value):
        if type(name) is not bytes:
            name = bytes(name, 'utf-8')
        loc = GL.glGetUniformLocation(self.handle, name)
        if not hasattr(value, '__len__'):
            GL.glUniform1i(loc, value)
        elif len(value) in range(1, 5):
            # Select the correct function
            { 1 : GL.glUniform1i,
              2 : GL.glUniform2i,
              3 : GL.glUniform3i,
              4 : GL.glUniform4i
              # Retrieve uniform location, and set it
            }[len(value)](loc, value)
        else:
            raise ValueError("Shader.setInt '{}' should be length 1-4 not {}"
                             .format(name, len(value)))


def compileProgram(vertexSource=None, fragmentSource=None):
    """Create and compile a vertex and fragment shader pair from their sources.

    Parameters
    ----------
    vertexSource, fragmentSource : str or list of str
        Vertex and fragment shader GLSL sources.

    Returns
    -------
    int
        Program object handle.

    """
    program = gltools.createProgramObjectARB()

    vertexShader = fragmentShader = None
    if vertexSource:
        vertexShader = gltools.compileShaderObjectARB(
            vertexSource, GL.GL_VERTEX_SHADER_ARB)
        gltools.attachObjectARB(program, vertexShader)
    if fragmentSource:
        fragmentShader = gltools.compileShaderObjectARB(
            fragmentSource, GL.GL_FRAGMENT_SHADER_ARB)
        gltools.attachObjectARB(program, fragmentShader)

    gltools.linkProgramObjectARB(program)
    # gltools.validateProgramARB(program)

    if vertexShader:
        gltools.detachObjectARB(program, vertexShader)
        gltools.deleteObjectARB(vertexShader)
    if fragmentShader:
        gltools.detachObjectARB(program, fragmentShader)
        gltools.deleteObjectARB(fragmentShader)

    return program


"""NOTE about frag shaders using FBO. If a floating point texture is being
used as a frame buffer (FBO object) then we should keep in the range -1:1
during frag shader. Otherwise we need to convert to 0:1. This means that
some shaders differ for FBO use if they're performing any signed math.
"""

fragFBOtoFrame = '''
    uniform sampler2D texture;

    float rand(vec2 seed){
        return fract(sin(dot(seed.xy ,vec2(12.9898,78.233))) * 43758.5453);
    }

    void main() {
        vec4 textureFrag = texture2D(texture,gl_TexCoord[0].st);
        gl_FragColor.rgb = textureFrag.rgb;
        //! if too high then show red/black noise
        if ( gl_FragColor.r>1.0 || gl_FragColor.g>1.0 || gl_FragColor.b>1.0) {
            gl_FragColor.rgb = vec3 (rand(gl_TexCoord[0].st), 0, 0);
        }
        //! if too low then show red/black noise
        else if ( gl_FragColor.r<0.0 || gl_FragColor.g<0.0 || gl_FragColor.b<0.0) {
            gl_FragColor.rgb = vec3 (0, 0, rand(gl_TexCoord[0].st));
        }
    }
    '''

# for stimuli with no texture (e.g. shapes)
fragSignedColor = '''
    void main() {
        gl_FragColor.rgb = ((gl_Color.rgb*2.0-1.0)+1.0)/2.0;
        gl_FragColor.a = gl_Color.a;
    }
    '''
fragSignedColor_adding = '''
    void main() {
        gl_FragColor.rgb = (gl_Color.rgb*2.0-1.0)/2.0;
        gl_FragColor.a = gl_Color.a;
    }
    '''
# for stimuli with just a colored texture
fragSignedColorTex = '''
    uniform sampler2D texture;
    void main() {
        vec4 textureFrag = texture2D(texture,gl_TexCoord[0].st);
        gl_FragColor.rgb = (textureFrag.rgb* (gl_Color.rgb*2.0-1.0)+1.0)/2.0;
        gl_FragColor.a = gl_Color.a*textureFrag.a;
    }
    '''
fragSignedColorTex_adding = '''
    uniform sampler2D texture;
    void main() {
        vec4 textureFrag = texture2D(texture,gl_TexCoord[0].st);
        gl_FragColor.rgb = textureFrag.rgb * (gl_Color.rgb*2.0-1.0)/2.0;
        gl_FragColor.a = gl_Color.a * textureFrag.a;
    }
    '''
# the shader for pyglet fonts doesn't use multitextures - just one texture
fragSignedColorTexFont = '''
    uniform sampler2D texture;
    uniform vec3 rgb;
    void main() {
        vec4 textureFrag = texture2D(texture,gl_TexCoord[0].st);
        gl_FragColor.rgb=rgb;
        gl_FragColor.a = gl_Color.a*textureFrag.a;
    }
    '''
# for stimuli with a colored texture and a mask (gratings, etc.)
fragSignedColorTexMask = '''
    uniform sampler2D texture, mask;
    void main() {
        vec4 textureFrag = texture2D(texture,gl_TexCoord[0].st);
        vec4 maskFrag = texture2D(mask,gl_TexCoord[1].st);
        gl_FragColor.a = gl_Color.a*maskFrag.a*textureFrag.a;
        gl_FragColor.rgb = (textureFrag.rgb* (gl_Color.rgb*2.0-1.0)+1.0)/2.0;
    }
    '''
fragSignedColorTexMask_adding = '''
    uniform sampler2D texture, mask;
    void main() {
        vec4 textureFrag = texture2D(texture,gl_TexCoord[0].st);
        vec4 maskFrag = texture2D(mask,gl_TexCoord[1].st);
        gl_FragColor.a = gl_Color.a * maskFrag.a * textureFrag.a;
        gl_FragColor.rgb = textureFrag.rgb * (gl_Color.rgb*2.0-1.0)/2.0;
    }
    '''
# RadialStim uses a 1D mask with a 2D texture
fragSignedColorTexMask1D = '''
    uniform sampler2D texture;
    uniform sampler1D mask;
    void main() {
        vec4 textureFrag = texture2D(texture,gl_TexCoord[0].st);
        vec4 maskFrag = texture1D(mask,gl_TexCoord[1].s);
        gl_FragColor.a = gl_Color.a*maskFrag.a*textureFrag.a;
        gl_FragColor.rgb = (textureFrag.rgb* (gl_Color.rgb*2.0-1.0)+1.0)/2.0;
    }
    '''
fragSignedColorTexMask1D_adding = '''
    uniform sampler2D texture;
    uniform sampler1D mask;
    void main() {
        vec4 textureFrag = texture2D(texture,gl_TexCoord[0].st);
        vec4 maskFrag = texture1D(mask,gl_TexCoord[1].s);
        gl_FragColor.a = gl_Color.a * maskFrag.a*textureFrag.a;
        gl_FragColor.rgb = textureFrag.rgb * (gl_Color.rgb*2.0-1.0)/2.0;
    }
    '''
# imageStim is providing its texture unsigned
fragImageStim = '''
    uniform sampler2D texture;
    uniform sampler2D mask;
    void main() {
        vec4 textureFrag = texture2D(texture,gl_TexCoord[0].st);
        vec4 maskFrag = texture2D(mask,gl_TexCoord[1].st);
        gl_FragColor.a = gl_Color.a*maskFrag.a*textureFrag.a;
        gl_FragColor.rgb = ((textureFrag.rgb*2.0-1.0)*(gl_Color.rgb*2.0-1.0)+1.0)/2.0;
    }
    '''
# imageStim is providing its texture unsigned
fragImageStim_adding = '''
    uniform sampler2D texture;
    uniform sampler2D mask;
    void main() {
        vec4 textureFrag = texture2D(texture,gl_TexCoord[0].st);
        vec4 maskFrag = texture2D(mask,gl_TexCoord[1].st);
        gl_FragColor.a = gl_Color.a*maskFrag.a*textureFrag.a;
        gl_FragColor.rgb = (textureFrag.rgb*2.0-1.0)*(gl_Color.rgb*2.0-1.0)/2.0;
    }
    '''
# in every case our vertex shader is simple (we don't transform coords)
vertSimple = """
    void main() {
            gl_FrontColor = gl_Color;
            gl_TexCoord[0] = gl_MultiTexCoord0;
            gl_TexCoord[1] = gl_MultiTexCoord1;
            gl_TexCoord[2] = gl_MultiTexCoord2;
            gl_Position =  ftransform();
    }
    """

vertPhongLighting = """
// Vertex shader for the Phong Shading Model
// 
// This code is based of the tutorial here:
//     https://www.opengl.org/sdk/docs/tutorials/ClockworkCoders/lighting.php
//
// Only supports directional and point light sources for now. Spotlights will be
// added later on.
//
#version 110
varying vec3 N;
varying vec3 v;
varying vec4 frontColor;

void main(void)  
{     
    v = vec3(gl_ModelViewMatrix * gl_Vertex);       
    N = normalize(gl_NormalMatrix * gl_Normal);
    
    gl_TexCoord[0] = gl_MultiTexCoord0;
    gl_TexCoord[1] = gl_MultiTexCoord1;
    gl_Position = ftransform();
    frontColor = gl_Color;
}
          
"""

fragPhongLighting = """
// Fragment shader for the Phong Shading Model
// 
// This code is based of the tutorial here:
//     https://www.opengl.org/sdk/docs/tutorials/ClockworkCoders/lighting.php
//
// Use `embedShaderSourceDefs` from gltools to enable the code path for diffuse 
// texture maps by setting DIFFUSE to 1. The number of lights can be specified 
// by setting MAX_LIGHTS, by default, the maximum should be 8. However, build
// your shader for the exact number of lights required. 
//
// Only supports directional and point light sources for now. Spotlights will be
// added later on.
//
#version 110
varying vec3 N;
varying vec3 v; 
varying vec4 frontColor;

#ifdef DIFFUSE_TEXTURE
    uniform sampler2D diffTexture;
#endif

// Calculate lighting attenuation using the same formula OpenGL uses
float calcAttenuation(float kConst, float kLinear, float kQuad, float dist) {
    return 1.0 / (kConst + kLinear * dist + kQuad * dist * dist);
}

void main (void)  
{  
#ifdef DIFFUSE_TEXTURE
    vec4 diffTexColor = texture2D(diffTexture, gl_TexCoord[0].st);
#endif 

#if MAX_LIGHTS > 0
    vec3 N = normalize(N);
    vec4 finalColor = vec4(0.0);
    // loop over available lights
    for (int i=0; i < MAX_LIGHTS; i++)
    {
        vec3 L;
        float attenuation = 1.0;  // default factor, no attenuation
        
        // check if directional, compute attenuation if a point source
        if (gl_LightSource[i].position.w == 0.0) 
        {
            // off at infinity, only use direction
            L = normalize(gl_LightSource[i].position.xyz);
            // attenuation is 1.0 (no attenuation for directional sources)
        } 
        else 
        {
            L = normalize(gl_LightSource[i].position.xyz - v);
            attenuation = calcAttenuation(
                gl_LightSource[i].constantAttenuation,
                gl_LightSource[i].linearAttenuation,
                gl_LightSource[i].quadraticAttenuation,
                length(gl_LightSource[i].position.xyz - v));
        }
        
        vec3 E = normalize(-v);
        vec3 R = normalize(-reflect(L, N)); 
        
        // combine scene ambient with object
        vec4 ambient = gl_FrontMaterial.diffuse * 
            (gl_FrontLightProduct[i].ambient + gl_LightModel.ambient); 
        
        // calculate diffuse component
        vec4 diffuse = gl_FrontLightProduct[i].diffuse * max(dot(N, L), 0.0);
#ifdef DIFFUSE_TEXTURE
        // multiply in material texture colors if specified
        diffuse *= diffTexColor;
        ambient *= diffTexColor;  // ambient should be modulated by diffuse color
#endif
        vec3 halfwayVec = normalize(L + E);
        vec4 specular = gl_FrontLightProduct[i].specular *
            pow(max(dot(N, halfwayVec), 0.0), gl_FrontMaterial.shininess);

        // clamp color values for specular and diffuse
        ambient = clamp(ambient, 0.0, 1.0); 
        diffuse = clamp(diffuse, 0.0, 1.0); 
        specular = clamp(specular, 0.0, 1.0); 
        
        // falloff with distance from eye? might be something to consider for 
        // realism
        vec4 emission = clamp(gl_FrontMaterial.emission, 0.0, 1.0);
        
        finalColor += ambient + emission + attenuation * (diffuse + specular);
    }
    gl_FragColor = finalColor;  // use texture alpha
#else
    // no lights, only track ambient component, frontColor modulates ambient
    vec4 ambient = gl_FrontLightProduct[0].ambient * gl_LightModel.ambient; 
    ambient = clamp(ambient, 0.0, 1.0); 
#ifdef DIFFUSE_TEXTURE
    gl_FragColor = ambient * texture2D(diffTexture, gl_TexCoord[0].st);
#else
    gl_FragColor = ambient;
#endif
#endif
}
"""

vertSkyBox = """
varying vec3 texCoord;
void main(void)  
{   
    texCoord = gl_Vertex;
    gl_Position = ftransform().xyww;
}      
"""

fragSkyBox = """
varying vec3 texCoord;
uniform samplerCube SkyTexture;
void main (void)  
{  
    gl_FragColor = texture(SkyTexture, texCoord);
}
"""

fragTextBox2 = '''
    uniform sampler2D texture;
    void main() {
        vec2 uv      = gl_TexCoord[0].xy;
        vec4 current = texture2D(texture, uv);

        float r = current.r;
        float g = current.g;
        float b = current.b;
        float a = current.a;
        gl_FragColor = vec4( gl_Color.rgb, (r+g+b)/2.);
    }
    '''
fragTextBox2alpha = '''
    uniform sampler2D texture;
    void main() {
        vec4 current = texture2D(texture,gl_TexCoord[0].st);
        gl_FragColor = vec4( gl_Color.rgb, current.a);
    }
    '''



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Created on Mon Jan 07 11:18:51 2013

@author: Sol
"""
from __future__ import absolute_import, print_function

from builtins import range
from builtins import object
import numpy as np
from weakref import proxy
from psychopy import core
from pyglet.gl import (glCallList, glGenLists, glNewList, glDisable, glEnable,
                       glTranslatef, glColor4f, glLineWidth, glBegin,
                       GL_LINES, glEndList, glDeleteLists, GL_COMPILE, glEnd,
                       GL_TEXTURE0, GL_TEXTURE_2D, GL_TEXTURE_ENV,
                       GL_TEXTURE_ENV_MODE, GL_MODULATE, GL_UNSIGNED_INT,
                       glPopMatrix, glBindTexture, glActiveTexture, glTexEnvf,
                       glPushMatrix, glCallLists, glVertex2i)
from . import parsedtext
getTime = core.getTime


class TextGrid(object):

    def __init__(self, text_box, line_color=None, line_width=1,
                 font_color=(1, 1, 1, 1), shape=None,
                 grid_horz_justification='left',
                 grid_vert_justification='top'):

        self._text_box = proxy(text_box)

        self._font_color = font_color
        if line_color:
            self._line_color = line_color
            self._line_width = line_width
        else:
            self._line_color = None
            self._line_width = None

        cfont = self._text_box._current_glfont
        max_size = cfont.max_tile_width, cfont.max_tile_height
        self._cell_size = (max_size[0],
                           max_size[1] + self._text_box._getPixelTextLineSpacing())
        if self._cell_size[0] == 0:
            print('ERROR WITH CELL SIZE!!!! ', self._text_box.getLabel())

        self._text_dlist = None
        self._gridlines_dlist = None

        self._text_document = None

        # Text Grid line_spacing
        te_size = [0, 0]
        if self._text_box._size:
            te_size = list(self._text_box._getPixelSize())

        if shape:
            self._shape = shape
        else:
            self._shape = (te_size[0] // self._cell_size[0],
                           te_size[1] // self._cell_size[1])

        self._size = (self._cell_size[0] * self._shape[0],
                      self._cell_size[1] * self._shape[1])
        resized = False
        if shape and self._size[0] > te_size[0]:
            te_size[0] = self._size[0]
            resized = True
        if shape and self._size[1] > te_size[1]:
            te_size[1] = self._size[1]
            resized = True
        if resized:
            self._text_box._setSize(te_size)
        # For now, The text grid is centered in the TextBox area.
        dx = (te_size[0] - self._size[0]) // 2
        dy = (te_size[1] - self._size[1]) // 2

        # TextGrid Position is position within the TextBox component.
        self._position = dx, dy

        # TextGrid cell boundaries
        self._col_lines = [int(np.floor(x)) for x in range(
            0, self._size[0] + 1, self._cell_size[0])]
        self._row_lines = [int(np.floor(y)) for y in range(
            0, -self._size[1] - 1, -self._cell_size[1])]

        self._apply_padding = False
        self._pad_top_proportion = 0
        self._pad_left_proportion = 0

        self.setHorzJust(grid_horz_justification)
        self.setVertJust(grid_vert_justification)

    def getSize(self):
        return self._size

    def getCellSize(self):
        return self._cell_size

    def getShape(self):
        return self._shape

    def getPosition(self):
        return self._position

    def getLineColor(self):
        return self._line_color

    def getLineWidth(self):
        return self._line_width

    def getHorzJust(self):
        return self._horz_justification

    def getVertJust(self):
        return self._vert_justification

    def setHorzJust(self, j):
        self._horz_justification = j
        self._pad_left_proportion = 0
        if j == 'center':
            self._pad_left_proportion = 0.5
        elif j == 'right':
            self._pad_left_proportion = 1.0

        self.applyPadding()

    def setVertJust(self, j):
        self._vert_justification = j
        self._pad_top_proportion = 0
        if j == 'center':
            self._pad_top_proportion = 0.5
        elif j == 'bottom':
            self._pad_top_proportion = 1.0

        self.applyPadding()

    def applyPadding(self):
        self._apply_padding = self._pad_left_proportion or (
            self._pad_top_proportion and self.getRowCountWithText() > 1)
        num_cols, num_rows = self._shape
        line_count = self.getRowCountWithText()
        for li in range(line_count):
            cline = self._text_document.getParsedLine(li)
            line_length = cline.getLength()
            if self._apply_padding:
                cline._trans_left = int(
                    (num_cols - line_length + 1) * self._pad_left_proportion)
                cline._trans_top = int(
                    (num_rows - line_count) * self._pad_top_proportion)
            else:
                cline._trans_left = 0
                cline._trans_top = 0

    def getRowCountWithText(self):
        if self._text_document:
            return min(self._shape[1], self._text_document.getParsedLineCount())
        return 0

    def _setText(self, text):
        self._text_document.deleteText(0, self._text_document.getTextLength(),
                                       text)

        self._deleteTextDL()
        self.applyPadding()
        return self._text_document.getDisplayedText()

    def setCurrentFontDisplayLists(self, dlists):
        self._current_font_display_lists = dlists

    def _deleteTextDL(self):
        if self._text_dlist:
            glDeleteLists(self._text_dlist, 1)
            self._text_dlist = 0

    def _deleteGridLinesDL(self):
        if self._gridlines_dlist:
            glDeleteLists(self._gridlines_dlist, 1)
            self._gridlines_dlist = None

    def _createParsedTextDocument(self, f):
        if self._shape:
            self._text_document = parsedtext.ParsedTextDocument(f, self)
            self._deleteTextDL()
            self.applyPadding()
        else:
            raise AttributeError(
                "Could not create _text_document. num_columns needs to be known.")

    def _text_glyphs_gl(self):
        if not self._text_dlist:
            dl_index = glGenLists(1)
            glNewList(dl_index, GL_COMPILE)

            # stime=getTime()

            # self._text_box._te_start_gl()

            ###
            glActiveTexture(GL_TEXTURE0)
            glEnable(GL_TEXTURE_2D)
            glBindTexture(
                GL_TEXTURE_2D, self._text_box._current_glfont.atlas.texid)
            glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE)
            glTranslatef(self._position[0], -self._position[1], 0)
            glPushMatrix()

            ###

            getLineInfoByIndex = self._text_document.getLineInfoByIndex
            active_text_style_dlist = self._current_font_display_lists.get
            cell_width, cell_height = self._cell_size
            num_cols, num_rows = self._shape
            line_spacing = self._text_box._getPixelTextLineSpacing()
            line_count = self.getRowCountWithText()

            glColor4f(*self._text_box._toRGBA(self._font_color))

            for r in range(line_count):
                cline, line_length, line_display_list, line_ords = getLineInfoByIndex(
                    r)
                if line_display_list[0] == 0:
                    line_display_list[0:line_length] = [
                        active_text_style_dlist(c) for c in line_ords]

                glTranslatef(cline._trans_left * cell_width, -
                             int(line_spacing/2.0 + cline._trans_top * cell_height), 0)
                glCallLists(line_length, GL_UNSIGNED_INT,
                            line_display_list[0:line_length].ctypes)
                cline._trans_left = 0
                glTranslatef(-line_length * cell_width - cline._trans_left * cell_width, -
                             cell_height + int(line_spacing/2.0 + cline._trans_top * cell_height), 0)

                ###
            glPopMatrix()
            glBindTexture(GL_TEXTURE_2D, 0)
            glDisable(GL_TEXTURE_2D)

            glEndList()
            self._text_dlist = dl_index
        glCallList(self._text_dlist)

    def _textgrid_lines_gl(self):
        if self._line_color:
            if not self._gridlines_dlist:
                dl_index = glGenLists(1)
                glNewList(dl_index, GL_COMPILE)

                glLineWidth(self._line_width)
                glColor4f(*self._text_box._toRGBA(self._line_color))
                glBegin(GL_LINES)
                for x in self._col_lines:
                    for y in self._row_lines:
                        if x == 0:
                            glVertex2i(x, y)
                            glVertex2i(int(self._size[0]), y)
                        if y == 0:
                            glVertex2i(x, y)
                            glVertex2i(x, int(-self._size[1]))
                glEnd()
                glColor4f(0.0, 0.0, 0.0, 1.0)
                glEndList()
                self._gridlines_dlist = dl_index
            glCallList(self._gridlines_dlist)

            # self._text_box._te_end_gl()

            # etime=getTime()

    def __del__(self):
        if self._text_dlist:
            glDeleteLists(self._text_dlist, 1)
            self._text_dlist = 0
        self._current_font_display_lists = None
        self._text_document._free()
        del self._text_document



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""This script is used to:
    - update the version numbers
    - update the psychopyVersions repo:
        - copy over the code
        - commit, tag and push(?)

    It should be run from the root of the main git repository, which should be
    next to a clone of the psychopy/versions git repository
"""
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

from builtins import input
import os, sys, shutil, subprocess
from os.path import join
from createInitFile import createInitFile
from psychopy.constants import PY3
from pathlib import Path

# MAIN is the root of the psychopy repo
MAIN = Path(__file__).parent.parent.parent.absolute()
# versions repo is next to MAIN
VERSIONS = MAIN.parent / 'versions'
print("Building release version from: ", MAIN)
print("To: ", VERSIONS)

if sys.platform == "darwin":
    gitgui = ["git", "gui"]
elif sys.platform == "linux":
    gitgui = ["cola"]
else:
    gitgui = ["git", "gui"]
    print("This script requires a unix-based terminal to run (for commands "
          "like `du -sck` to work)")
    sys.exit()

def getSHA(cwd='.'):
    if cwd == '.':
        cwd = os.getcwd()
    # get the SHA of the git HEAD
    SHA_string = subprocess.check_output(
        ['git', 'rev-parse', '--short', 'HEAD'],
        cwd=cwd).split()[0]
    if PY3:
        SHA_string = SHA_string.decode('utf-8')

    # convert to hex from a string and return it
    print('SHA:', SHA_string, 'for repo:', cwd)
    return SHA_string


def buildRelease(versionStr, noCommit=False, interactive=True):
    #
    createInitFile(dist='sdist', version=versionStr, sha=getSHA())
    dest = join(VERSIONS, "psychopy")
    shutil.rmtree(dest)
    ignores = shutil.ignore_patterns("demos", "docs", "tests", "pylink",
                                     "*.pyo", "*.pyc", "*.orig", "*.bak",
                                     ".DS_Store", ".coverage")
    shutil.copytree("psychopy", dest, symlinks=False, ignore=ignores)

    # todo: would be nice to check here that we didn't accidentally add anything large (check new folder size)
    Mb = float(subprocess.check_output(["du", "-sck", dest]).split()[0])/10**3
    print("size for '%s' will be: %.2f Mb" %(versionStr, Mb))
    if noCommit:
        return False

    if interactive:
        ok = input("OK to continue? [n]y :")
        if ok != "y":
            return False

    lastSHA = getSHA(cwd=VERSIONS)
    print('updating: git add --all')
    output = subprocess.check_output(["git", "add", "--all"], cwd=VERSIONS)
    if interactive:
        ok = subprocess.call(gitgui, cwd=VERSIONS)
        if lastSHA == getSHA():
            # we didn't commit the changes so quit
            print("no git commit was made: exiting")
            return False
    else:
        print("committing: git commit -m 'release version %s'" %versionStr)
        subprocess.call(
            ["git", "commit", "-m", "'release version %s'" %versionStr],
            cwd=VERSIONS)

    print("tagging: git tag -m 'release %s'" %versionStr)
    ok = subprocess.call(
        ["git", "tag", versionStr, "-m", "'release %s'" %versionStr],
        cwd=VERSIONS)

    print("'versions' tags are now:", subprocess.check_output(
        ["git","tag"], cwd=VERSIONS).split())
    print('pushing: git push origin %s' %versionStr)
    output = subprocess.check_output(["git", "push", "origin", "%s" % versionStr],
                         cwd=VERSIONS)
    print(output)

    # revert the __init__ file to non-ditribution state
    print('reverting the main master branch: git checkout HEAD psychopy/__init__.py ')
    print(subprocess.check_output(
         ["git", "checkout", "HEAD", "psychopy/__init__.py"],
         cwd=MAIN))
    return True  # success


if __name__ == "__main__":
    if "--noCommit" in sys.argv:
        noCommit = True
    else:
        noCommit = False
    if "--noInteractive" not in sys.argv:
        interactive = True
    else:
        interactive = False
    # todo: update versions first
    versionStr = input("version:")
    buildRelease(versionStr, noCommit=noCommit, interactive=interactive)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Demo of gabor dots, using autodraw.
"""

from __future__ import division

from psychopy import visual, core, event

# Create a window to draw in
win = visual.Window((600, 600), allowGUI=False,
                      monitor='testMonitor', units='deg')

# Initialize
gabor_shape = visual.GratingStim(win, mask='gauss', sf=3)
dotPatch = visual.DotStim(win, color='black',
    dotLife=5,  # lifetime of a dot in frames (if this is long density artifacts can occur in the stimulus)
    signalDots='different',  # are the signal and noise dots 'different' or 'same' popns (see Scase et al)
    noiseDots='direction',  # do the noise dots follow random- 'walk', 'direction', or 'position'
    fieldPos=[0.0, 0.0], nDots=40, fieldSize=3,
    speed=0.05, fieldShape='circle', coherence=0.5,
    element=gabor_shape, name='dotPatch')
message = visual.TextStim(win, text='Any key to quit', pos=(0, -5))

# always draw
dotPatch.autoDraw = True
message.autoDraw = True

while not event.getKeys():
    win.flip()  # redraw the buffer, autodraw does the rest

win.close()
core.quit()

# The contents of this file are in the public domain.



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import print_function
import sys, os
import subprocess

thisDir,filename = os.path.split(os.path.abspath(__file__))
os.chdir(thisDir)

argv = sys.argv

if argv[0].endswith('run.py'):#that's this script
    argv.pop(0)  # remove run.py

#create extra args and then append main args to them
#if the user specifies a directory that should come last
extraArgs = []

#if user didn't specify a traceback deetail level then set to be short
overrideTraceBack=True
for thisArg in argv:
    if thisArg.startswith('--tb='):
        overrideTraceBack=False
if overrideTraceBack:
    extraArgs.append('--tb=short')

#always add doctests if not included
if '--doctest-modules' not in argv:
    extraArgs.append('--doctest-modules') #doctests
#add coverage if requested
if 'cover' in argv:
    argv.remove('cover')
    extraArgs.extend(['--cov','psychopy'])

#use the pytest module itself if available
#otherwise use the precompiled pytest script
try:
    import pytest
    havePytest=True
except ImportError:
    havePytest=False

extraArgs.extend(argv)
if havePytest:
    command = 'py.test ' + ' '.join(extraArgs)
else:
    command = '%s -u runPytest.py ' %(sys.executable)
    command = command + ' ' + ' '.join(extraArgs)
print(command)
subprocess.Popen(command, shell=True)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Functions and classes related to code folding with Scintilla."""

# Part of the PsychoPy library
# Copyright (C) 2002-2018 Jonathan Peirce (C) 2019-2020 Open Science Tools Ltd.
# Distributed under the terms of the GNU General Public License (GPL).

import wx.stc


class CodeEditorFoldingMixin(object):
    """Mixin for adding code folding functionality to a CodeEditor control.

    """
    def OnMarginClick(self, evt):
        """Event for when a margin is clicked."""
        # fold and unfold as needed
        if evt.GetMargin() == 1:
            if evt.GetShift() and evt.GetControl():
                self.FoldAll()
            else:
                lineClicked = self.LineFromPosition(evt.GetPosition())
                _flag = wx.stc.STC_FOLDLEVELHEADERFLAG
                if self.GetFoldLevel(lineClicked) & _flag:
                    if evt.GetShift():
                        self.SetFoldExpanded(lineClicked, True)
                        self.Expand(lineClicked, True, True, 1)
                    elif evt.GetControl():
                        if self.GetFoldExpanded(lineClicked):
                            self.SetFoldExpanded(lineClicked, False)
                            self.Expand(lineClicked, False, True, 0)
                        else:
                            self.SetFoldExpanded(lineClicked, True)
                            self.Expand(lineClicked, True, True, 100)
                    else:
                        self.ToggleFold(lineClicked)

    def FoldAll(self):
        """Fold all code blocks."""
        lineCount = self.GetLineCount()
        expanding = True

        # find out if we are folding or unfolding
        for lineNum in range(lineCount):
            if self.GetFoldLevel(lineNum) & wx.stc.STC_FOLDLEVELHEADERFLAG:
                expanding = not self.GetFoldExpanded(lineNum)
                break

        lineNum = 0
        _flag = wx.stc.STC_FOLDLEVELHEADERFLAG
        _mask = wx.stc.STC_FOLDLEVELNUMBERMASK
        _base = wx.stc.STC_FOLDLEVELBASE
        while lineNum < lineCount:
            level = self.GetFoldLevel(lineNum)
            if level & _flag and level & _mask == _base:
                if expanding:
                    self.SetFoldExpanded(lineNum, True)
                    lineNum = self.Expand(lineNum, True)
                    lineNum -= 1
                else:
                    lastChild = self.GetLastChild(lineNum, -1)
                    self.SetFoldExpanded(lineNum, False)
                    if lastChild > lineNum:
                        self.HideLines(lineNum + 1, lastChild)
            lineNum += 1

    def Expand(self, line, doExpand, force=False, visLevels=0, level=-1):
        lastChild = self.GetLastChild(line, level)
        line += 1

        while line <= lastChild:
            if force:
                if visLevels > 0:
                    self.ShowLines(line, line)
                else:
                    self.HideLines(line, line)
            else:
                if doExpand:
                    self.ShowLines(line, line)
            if level == -1:
                level = self.GetFoldLevel(line)
            if level & wx.stc.STC_FOLDLEVELHEADERFLAG:
                if force:
                    if visLevels > 1:
                        self.SetFoldExpanded(line, True)
                    else:
                        self.SetFoldExpanded(line, False)
                    line = self.Expand(line, doExpand, force, visLevels - 1)
                else:
                    if doExpand and self.GetFoldExpanded(line):
                        line = self.Expand(line, True, force, visLevels - 1)
                    else:
                        line = self.Expand(line, False, force, visLevels - 1)
            else:
                line += 1
        return line


def getFolds(doc):
    """Get the lines and levels of folds found by the Scintilla. This can be
    used to create parsers for the source tree among other things.

    Parameters
    ----------
    doc : wx.stc.StyledText
        Styled text object.

    Returns
    -------
    list
        List of fold levels and line numbers.

    """
    # Go over file and get all the folds.
    foldLines = []
    for lineno in range(doc.GetLineCount()):
        foldLevelFlags = doc.GetFoldLevel(lineno)
        foldLevel = \
            (foldLevelFlags & wx.stc.STC_FOLDLEVELNUMBERMASK) - \
            wx.stc.STC_FOLDLEVELBASE  # offset
        isFoldStart = (foldLevelFlags & wx.stc.STC_FOLDLEVELHEADERFLAG) > 0

        if isFoldStart:
            foldLines.append((foldLevel, lineno))

    return foldLines



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
A stimulus class for playing movies (mpeg, avi, etc...) in PsychoPy
using VLC.  movie4 does /not/ require avbin to be installed.

Testing has only been done on Windows.

VlcMovieStim requires:
~~~~~~~~~~~~~~~~~~~~~

1. VLC. Just install the standard VLC of the same bitness as python
    for your OS.

    http://www.videolan.org/vlc/index.html

2. pip install python-vlc

To play a video:
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

Create a new instance, `mov = visual.VlcMovieStim(..)`

    shouldflip = mov.play()
    continueRoutine = True
    while continueRoutine:
        if shouldflip:
            # Draw other stimuli here
            win.flip()
        else:
            time.sleep(0.001)
        shouldflip = mov.draw()


To fix "stale cache" VLC errors on Windows...
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

Libvlc has a bug related to changing from daylight standard time to daylight 
saving time. After the time change libvlc complains that the plugin cache is 
stale. The error message begins with "main libvlc error: stale plugins cache: 
...". VLC provides an executable to update the plugin cache timestamps. To run 
it, execute this as Administrator:

    cd "C:\Program Files\VideoLAN\VLC"
    vlc-cache-gen.exe "C:\Program Files\VideoLAN\VLC\plugins"


Not implemented:
~~~~~~~~~~~~~~~~

* Horizontal/vertical flipping
* getCurrentFrameTime
* getPercentageComplete

"""

# Part of the PsychoPy library
# Copyright (C) 2002-2018 Jonathan Peirce (C) 2019-2020 Open Science Tools Ltd.
# Distributed under the terms of the GNU General Public License (GPL).
#
# VlcMovieStim contributed by Dan Fitch, April 2019.
# The MovieStim2 class was taken and rewritten to use only vlc

from __future__ import absolute_import, division, print_function

# If True then, on each flip a new movie frame is displayed, the frame index,
# flip time, and time since last movie frame flip will be printed
reportNDroppedFrames = 10

import os
import sys
import threading
import weakref

# Ensure setting pyglet.options['debug_gl'] to False is done prior to any
# other calls to pyglet or pyglet submodules, otherwise it may not get picked
# up by the pyglet GL engine and have no effect.
# Shaders will work but require OpenGL2.0 drivers AND PyOpenGL3.0+
import pyglet
pyglet.options['debug_gl'] = False
GL = pyglet.gl


import psychopy
from psychopy import core, logging

from psychopy.tools.arraytools import val2array
from psychopy.tools.attributetools import logAttrib, setAttribute
from psychopy.tools.filetools import pathToString
from psychopy.visual.basevisual import BaseVisualStim, ContainerMixin
from psychopy.clock import Clock
from psychopy.constants import FINISHED, NOT_STARTED, PAUSED, PLAYING, STOPPED
from psychopy.tools.monitorunittools import convertToPix

import ctypes
import numpy

try:
    import vlc
except Exception as err:
    if sys.maxsize == 9223372036854775807:
        bits = 64
    else:
        bits = 32
    if "wrong architecture" in err:
        msg = ("Failed to import vlc module for MovieStim2.\n"
               "You're using %i-bit python. Is your VLC install the same?"
               % bits)
        raise OSError(msg)
    else:
        raise err



class TexturedRect:
    def __init__(self, texture_id, opacity=1.0):
        self.texture_id = texture_id
        self.pos = (0.0, 0.0)
        self.size = (1.0, 1.0)
        self.angle = 0
        self.opacity = opacity
        self.init_vertexes()

    def init_vertexes(self):
        x = 0.5
        y = 0.5
        self.vertex_list = pyglet.graphics.vertex_list(4, ('v2f', [-x,y, x,y, -x,-y, x,-y]), ('t2f', [0,0, 1,0, 0,1, 1,1]))

    def set_position_and_size(self, pos, size):
        self.pos = pos
        self.size = size

    def draw(self):
        GL.glPushMatrix()
        GL.glRotatef(self.angle, 0, 0, 1)
        GL.glTranslatef(self.pos[0], self.pos[1], 0)
        GL.glScalef(self.size[0], self.size[1], 1)
        GL.glColor4f(1,1,1,1)
        GL.glActiveTexture(GL.GL_TEXTURE0)
        GL.glEnable(GL.GL_TEXTURE_2D)
        GL.glBindTexture(GL.GL_TEXTURE_2D, self.texture_id)
        self.vertex_list.draw(GL.GL_TRIANGLE_STRIP)
        GL.glDisable(GL.GL_TEXTURE_2D)
        GL.glPopMatrix()

# vlc.CallbackDecorators in python-vlc lib are incorrect and don't match VLC docs
CorrectVideoLockCb = ctypes.CFUNCTYPE(ctypes.c_void_p, ctypes.c_void_p, ctypes.POINTER(ctypes.c_void_p))
CorrectVideoUnlockCb = ctypes.CFUNCTYPE(ctypes.c_void_p, ctypes.c_void_p, ctypes.c_void_p, ctypes.POINTER(ctypes.c_void_p))

@CorrectVideoLockCb
def vlcLockCallback(user_data, planes):
    self = ctypes.cast(user_data, ctypes.POINTER(ctypes.py_object)).contents.value
    self.pixel_lock.acquire()
    # Tell VLC to take the data and stuff it into the buffer
    planes[0] = ctypes.cast(self.pixel_buffer, ctypes.c_void_p)

@CorrectVideoUnlockCb
def vlcUnlockCallback(user_data, picture, planes):
    self = ctypes.cast(user_data, ctypes.POINTER(ctypes.py_object)).contents.value
    self.pixel_lock.release()

@vlc.CallbackDecorators.VideoDisplayCb
def vlcDisplayCallback(user_data, picture):
    self = ctypes.cast(user_data, ctypes.POINTER(ctypes.py_object)).contents.value
    self.frame_counter += 1

def vlcTimeCallback(event, user_data, player):
    if user_data():
        tm = -event.u.new_time/1000.0
        user_data()._vlc_clock.reset(tm)
    return

def vlcEndReached(event, user_data, player):
    if user_data():
        user_data()._onEos()
    return


class VlcMovieStim(BaseVisualStim, ContainerMixin):
    """A stimulus class for playing movies (mpeg, avi, etc...) in PsychoPy
    that uses VLC and does not require avbin. The VLC media player must be 
    installed on the psychopy computer.
    """

    def __init__(self, win,
                 filename="",
                 units='pix',
                 size=None,
                 pos=(0.0, 0.0),
                 ori=0.0,
                 flipVert=False,
                 flipHoriz=False,
                 color=(1.0, 1.0, 1.0),
                 colorSpace='rgb',
                 opacity=1.0,
                 volume=1.0,
                 name='',
                 loop=False,
                 autoLog=True,
                 depth=0.0,
                 noAudio=False,
                 vframe_callback=None,
                 fps=None,
                 interpolate=True):
        """
        :Parameters:

            filename :
                a string giving the relative or absolute path to the movie.
            flipVert : True or *False*
                If True then the movie will be top-bottom flipped
            flipHoriz : True or *False*
                If True then the movie will be right-left flipped
            volume :
                The nominal level is 100, and 0 is silence.
            loop : bool, optional
                Whether to start the movie over from the beginning if draw is
                called and the movie is done.

        """
        # what local vars are defined (these are the init params) for use
        # by __repr__
        self._initParams = dir()
        self._initParams.remove('self')
        super(VlcMovieStim, self).__init__(win, units=units, name=name,
                                         autoLog=False)
        # check for pyglet
        if win.winType != 'pyglet':
            logging.error('Movie stimuli can only be used with a pyglet window')
            core.quit()
        self._retracerate = win._monitorFrameRate
        if self._retracerate is None:
            self._retracerate = win.getActualFrameRate()
        if self._retracerate is None:
            logging.warning("FrameRate could not be supplied by psychopy; "
                            "defaulting to 60.0")
            self._retracerate = 60.0
        self.filename = pathToString(filename)
        self.loop = loop
        self.flipVert = flipVert
        self.flipHoriz = flipHoriz
        self.pos = numpy.asarray(pos, float)
        self.size = numpy.asarray(size, float)
        self.depth = depth
        self.opacity = float(opacity)
        self.volume = volume
        self.no_audio = noAudio

        self.interpolate = interpolate
        self._texture_id = GL.GLuint()
        GL.glGenTextures(1, ctypes.byref(self._texture_id))

        self._pause_time = 0
        self._vlc_clock = Clock()
        self._vlc_initialized = False
        self._reset()
        self.loadMovie(self.filename)
        self.setVolume(volume)
        self.nDroppedFrames = 0

        self.ori = ori
        # set autoLog (now that params have been initialised)
        self.autoLog = autoLog
        if autoLog:
            logging.exp("Created {} = {}".format(self.name, self))

    def _reset(self):
        self.frame_counter = 0
        self.current_frame = 0
        self.duration = None
        self.status = NOT_STARTED
        self.width = None
        self.height = None
        self.frame_rate = None

        if self._vlc_initialized:
            self._release_vlc()

    def setMovie(self, filename, log=True):
        """See `~MovieStim.loadMovie` (the functions are identical).

        This form is provided for syntactic consistency with other
        visual stimuli.
        """
        self.loadMovie(filename, log=log)

    def loadMovie(self, filename, log=True):
        """Load a movie from file

        :Parameters:

            filename: string
                The name of the file, including path if necessary

        Due to VLC oddness, .duration is not correct until the movie starts playing.
        """
        self._reset()
        self.filename = pathToString(filename)

        # Initialize VLC
        self._vlc_start()

        self.status = NOT_STARTED
        logAttrib(self, log, 'movie', filename)

    def _vlc_start(self):
        """
        Create the vlc stream player for the video using python-vlc.
        """
        if not os.access(self.filename, os.R_OK):
            raise RuntimeError('Error: %s file not readable' % self.filename)
        if self.no_audio:
            instance = vlc.Instance("--no-audio")
        else:
            instance = vlc.Instance()
        try:
            stream = instance.media_new(self.filename)
        except NameError:
            msg = 'NameError: %s vs LibVLC %s'
            raise ImportError(msg % (vlc.__version__,
                                     vlc.libvlc_get_version()))

        player = instance.media_player_new()
        player.set_media(stream)

        # Load up the file
        stream.parse()
        size = player.video_get_size()
        self.video_width = size[0]
        self.video_height = size[1]
        self.frame_rate = player.get_fps()
        self.frame_counter = 0

        # TODO: Why is duration -1 still even after parsing? Newer vlc docs seem to hint this won't work until playback starts
        duration = player.get_length()
        logging.warning("Video is %ix%i, duration %s, fps %s" % (self.video_width, self.video_height, duration, self.frame_rate))
        logging.flush()

        # We assume we can use the RGBA format here
        player.video_set_format("RGBA", self.video_width, self.video_height, self.video_width << 2)

        # Configure a lock and a buffer for the pixels coming from VLC
        self.pixel_lock = threading.Lock()
        self.pixel_buffer = (ctypes.c_ubyte * self.video_width * self.video_height * 4)()

        # Once you set these callbacks, you are in complete control of what to do with the video buffer
        selfref = ctypes.cast(ctypes.pointer(ctypes.py_object(self)), ctypes.c_void_p)
        player.video_set_callbacks(vlcLockCallback, vlcUnlockCallback, vlcDisplayCallback, selfref)

        manager = player.event_manager()
        manager.event_attach(
            vlc.EventType.MediaPlayerTimeChanged, vlcTimeCallback,
            weakref.ref(self), player)
        manager.event_attach(
            vlc.EventType.MediaPlayerEndReached, vlcEndReached,
            weakref.ref(self), player)

        # Keep references
        self._self_ref = selfref
        self._instance = instance
        self._player = player
        self._stream = stream
        self._manager = manager

        logging.info("Initialized VLC...")
        self._vlc_initialized = True

    def _release_vlc(self):
        logging.info("Releasing VLC...")

        if self._manager: self._manager.event_detach(vlc.EventType.MediaPlayerTimeChanged)
        if self._player: self._player.stop()
        if self._stream: self._stream.release()
        if self._instance: self._instance.release()

        self._stream = None
        self._stream_event_manager = None
        self._player = None
        self._instance = None
        self._vlc_initialized = False

    def _update_texture(self):
        """
        Take the pixel buffer (assumed to be RGBA)
        and cram it into the GL texture
        """
        with self.pixel_lock:
            GL.glEnable(GL.GL_TEXTURE_2D)
            GL.glBindTexture(GL.GL_TEXTURE_2D, self._texture_id)
            GL.glPixelStorei(GL.GL_UNPACK_ALIGNMENT, 1)
            if self.interpolate:
                interpolation = GL.GL_LINEAR
            else:
                interpolation = GL.GL_NEAREST
            GL.glTexParameteri(GL.GL_TEXTURE_2D, GL.GL_TEXTURE_MAG_FILTER, interpolation)
            GL.glTexParameteri(GL.GL_TEXTURE_2D, GL.GL_TEXTURE_MIN_FILTER, interpolation)
            GL.glTexImage2D(GL.GL_TEXTURE_2D, 0, GL.GL_RGB,
                            self.video_width,
                            self.video_height,
                            0, GL.GL_RGBA, GL.GL_UNSIGNED_BYTE,
                            self.pixel_buffer)
            GL.glDisable(GL.GL_TEXTURE_2D)


    def play(self, log=True):
        """Start or continue a paused movie from current position.
        """
        cstat = self.status
        if cstat != PLAYING:
            self.status = PLAYING

            if self._pause_time:
                self._vlc_clock.reset(self._pause_time)

            if self._player:
                if cstat == PAUSED:
                    self._player.pause()
                else:
                    self._player.play()

            if log and self.autoLog:
                self.win.logOnFlip("Set %s playing" % (self.name),
                                   level=logging.EXP, obj=self)

            self._update_texture()
            return self.current_frame

    def pause(self, log=True):
        """Pause the current point in the movie.
        """
        if self.status == PLAYING:
            self.status = PAUSED
            player = self._player
            if player and player.can_pause():
                player.pause()
            if log and self.autoLog:
                self.win.logOnFlip("Set %s paused" % self.name,
                                   level=logging.EXP, obj=self)
            self._pause_time = self._vlc_clock.getTime()
            return True
        if log and self.autoLog:
            self.win.logOnFlip("Failed Set %s paused" % self.name,
                               level=logging.EXP, obj=self)
        return False

    def stop(self, log=True):
        """Stop the current point in the movie (sound will stop,
        current frame will not advance). Once stopped the movie cannot
        be restarted - it must be loaded again.

        Use pause() if you may need to restart the movie.
        """
        if self.status != STOPPED:
            self.status = STOPPED
            self._reset()
            if log and self.autoLog:
                self.win.logOnFlip("Set %s stopped" % (self.name),
                                   level=logging.EXP, obj=self)

    def seek(self, timestamp, log=True):
        """Seek to a particular timestamp in the movie.
        """
        if self.status in [PLAYING, PAUSED]:
            player = self._player
            if player and player.is_seekable():
                player.set_time(int(timestamp * 1000.0))
                self._vlc_clock.reset(timestamp)

                if self.status == PAUSED:
                    self._pause_time = timestamp

            if log:
                logAttrib(self, log, 'seek', timestamp)

    def setFlipHoriz(self, newVal=True, log=True):
        """If set to True then the movie will be flipped horizontally
        (left-to-right). Note that this is relative to the original,
        not relative to the current state.
        """
        self.flipHoriz = newVal
        logAttrib(self, log, 'flipHoriz')

    def setFlipVert(self, newVal=True, log=True):
        """If set to True then the movie will be flipped vertically
        (top-to-bottom). Note that this is relative to the original,
        not relative to the current state.
        """
        self.flipVert = not newVal
        logAttrib(self, log, 'flipVert')

    def setVolume(self, v):
        """Set the audio track volume. 0 = mute, 100 = 0 dB. float values
        between 0.0 and 1.0 are also accepted, and scaled to an int
        between 0 and 100.
        """
        if self._player:
            if 0.0 <= v <= 1.0 and isinstance(v, float):
                v = int(v * 100)
            else:
                v = int(v)
            self.volume = v
            if self._player:
                self._player.audio_set_volume(v)

    def getVolume(self):
        """Returns the current movie audio volume.

        0 is no audio, 100 is max audio volume.
        """
        if self._player:
            self.volume = self._player.audio_get_volume()
        return self.volume

    def getFPS(self):
        """
        Returns the movie frames per second playback speed.
        """
        return self.frame_rate

    def getCurrentFrameNumber(self):
        """Get the current movie frame number.
        The first frame number in a file is 1.
        """
        return self.frame_counter

    def getCurrentFrameTime(self):
        """Get the time that the movie file specified the current
        video frame as having.
        """
        return self._vlc_clock.getTime()

    def getPercentageComplete(self):
        """Provides a value between 0.0 and 100.0, indicating the
        amount of the movie that has been already played.
        """
        return self._player.get_position() * 100.0

    def _draw_rectangle(self, win):
        # make sure that textures are on and GL_TEXTURE0 is active
        GL.glActiveTexture(GL.GL_TEXTURE0)
        GL.glEnable(GL.GL_TEXTURE_2D)
        # sets opacity (1,1,1 = RGB placeholder)
        GL.glColor4f(1, 1, 1, self.opacity)
        GL.glPushMatrix()
        self.win.setScale('pix')
        # move to centre of stimulus and rotate
        vertsPix = self.verticesPix

        array = (GL.GLfloat * 32)(
            1, 1,  # texture coords
            vertsPix[0, 0], vertsPix[0, 1], 0.,  # vertex
            0, 1,
            vertsPix[1, 0], vertsPix[1, 1], 0.,
            0, 0,
            vertsPix[2, 0], vertsPix[2, 1], 0.,
            1, 0,
            vertsPix[3, 0], vertsPix[3, 1], 0.,
        )
        GL.glPushAttrib(GL.GL_ENABLE_BIT)
        GL.glEnable(GL.GL_TEXTURE_2D)
        GL.glBindTexture(GL.GL_TEXTURE_2D, self._texture_id)
        GL.glPushClientAttrib(GL.GL_CLIENT_VERTEX_ARRAY_BIT)
        # 2D texture array, 3D vertex array
        GL.glInterleavedArrays(GL.GL_T2F_V3F, 0, array)
        GL.glDrawArrays(GL.GL_QUADS, 0, 4)
        GL.glPopClientAttrib()
        GL.glPopAttrib()
        GL.glPopMatrix()

    def draw(self, win=None):
        """Draw the current frame to a particular visual.Window (or to the
        default win for this object if not specified).
        The current position in the movie will be determined automatically.

        This method should be called on every frame that the movie is meant
        to appear.
        """
        if self.status == NOT_STARTED or (self.status == FINISHED and self.loop):
            self.play()
        elif self.status == FINISHED and not self.loop:
            return
        if win is None:
            win = self.win
        self._selectWindow(win)

        self._update_texture()
        self._draw_rectangle(win)

        if self.current_frame != self.frame_counter:
            self.current_frame = self.frame_counter
            return True

    def setContrast(self):
        """Not yet implemented
        """
        pass

    def _unload(self):
        if self._vlc_initialized:
            self._release_vlc()
        if self._texture_id is not None:
            GL.glDeleteTextures(1, self._texture_id)
            self._texture_id = None
        self.status = FINISHED

    def _onEos(self):
        if self.loop:
            self.seek(0.0)
        else:
            self.status = FINISHED
            self.stop()
        if self.autoLog:
            self.win.logOnFlip("Set %s finished" % self.name,
                               level=logging.EXP, obj=self)

    def __del__(self):
        try:
            self._unload()
        except ModuleNotFoundError:
            pass  # has probably been garbage-collected already

    def setAutoDraw(self, val, log=None):
        """Add or remove a stimulus from the list of stimuli that will be
        automatically drawn on each flip

        :parameters:
            - val: True/False
                True to add the stimulus to the draw list, False to remove it
        """
        if val:
            self.play(log=False)  # set to play in case stopped
        else:
            self.pause(log=False)
        # add to drawing list and update status
        setAttribute(self, 'autoDraw', val, log)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import absolute_import, print_function

import wx
import IPython.gui.wx.ipython_view


class ShellFrame(wx.Frame):
    def __init__(self, parent, ID, title, files=None, app=None):
        self.app = app
        self.frameType = 'shell'
        # things the user doesn't set like winsize etc
        self.appData = self.app.prefs.appData['coder']
        self.prefs = self.app.prefs.coder
        self.appPrefs = self.app.prefs.app
        self.paths = self.app.prefs.paths
        self.IDs = self.app.IDs

        wx.Frame.__init__(self, parent, ID, title, size=((600, 400)))
        self.sizer = wx.BoxSizer()

        IPShell = IPython.gui.wx.ipython_view.IPShellWidget
        self.ipython = IPShell(parent=self, background_color='WHITE',)
        # turn off threading - interferes with pygame thread
        self.ipython.options['threading']['value'] = 'False'
        tracker = self.ipython.updateOptionTracker
        self.ipython.IP.set_threading(False)
        self.ipython.threading_option.SetValue(False)
        # allow a write fmethod for the window
        self.ipython.cout.write = self.ipython.text_ctrl.write
        self.ipython.write = self.ipython.text_ctrl.write

        # set background to white
        # this setting isn't used by __init__ apparently
        self.ipython.options['background_color']['value'] = 'WHITE'
        self.ipython.text_ctrl.setBackgroundColor(
            self.ipython.options['background_color']['value'])
        self.ipython.background_option.SetValue(True)
        tracker('background_color',
                self.ipython.options['background_color']['value'])
        # scintilla autocompletion method
        self.ipython.completion_option.SetValue(True)
        self.ipython.options['completion']['value'] = 'STC'
        self.ipython.text_ctrl.setCompletionMethod(
            self.ipython.options['completion']['value'])
        tracker('completion', self.ipython.options['completion']['value'])
        self.ipython.text_ctrl.SetFocus()

        self.sizer.Add(self.ipython, 0, wx.EXPAND)
        self.SendSizeEvent()

        self.SetSizerAndFit(self.sizer)
        self.SetAutoLayout(1)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Part of the PsychoPy library
# Copyright (C) 2002-2018 Jonathan Peirce (C) 2019-2020 Open Science Tools Ltd.
# Distributed under the terms of the GNU General Public License (GPL).

"""Functions and classes related to attribute handling
"""
from __future__ import absolute_import, division, print_function

from past.builtins import basestring
from past.utils import old_div
from builtins import object
import numpy

from psychopy import logging


class attributeSetter(object):
    """Makes functions appear as attributes. Takes care of autologging.
    """

    def __init__(self, func, doc=None):
        self.func = func
        if doc is not None:
            self.__doc__ = doc
        else:
            self.__doc__ = func.__doc__

    def __set__(self, obj, value):
        newValue = self.func(obj, value)
        # log=None defaults to obj.autoLog:
        logAttrib(obj, log=None, attrib=self.func.__name__,
                  value=value)

        # Useful for inspection/debugging. Keeps track of all settings of
        # attributes.
        '''
        import traceback
        origin = traceback.extract_stack()[-2]
        # print('%s.%s = %s (line %i)' %(obj.__class__.__name__,
        #        self.func.__name__, value.__repr__(), origin[1]))  # short
        # print('%s.%s = %s (%s in %s, line %i' %(obj.__class__.__name__,
        #        self.func.__name__, value.__repr__(),
        #        origin[1], origin[0].split('/')[-1],
        #        origin[1], origin[3].__repr__())))  # long
        '''
        return newValue

    def __repr__(self):
        return repr(self.__getattribute__)


def setAttribute(self, attrib, value, log,
                 operation=False, stealth=False):
    """This function is useful to direct the old set* functions to the
    @attributeSetter.

    It has the same functionality but supports logging control as well, making
    it useful for cross-@attributeSetter calls as well with log=False.

    Typical usage: e.g. in setSize(self, value, operation, log=None) do::

        def setSize(self, value, operation, log=None):
            # call attributeSetter:
            setAttribute(self, 'size', value, log, operation)

    Sets an object property (scalar or numpy array), optionally with
    an operation given in a string. If operation is None or '', value
    is multiplied with old to keep shape.

    If `stealth` is True, then use self.__dict[key] = value and avoid
    calling attributeSetters. If `stealth` is False, use `setattr()`.
    `autoLog`  controls the value of autoLog during this `setattr()`.

    History: introduced in version 1.79 to avoid exec-calls.
    Even though it looks complex, it is very fast :-)
    """

    # Change the value of "value" if there is an operation. Even if it is '',
    # which indicates that this value could potentially be subjected to an
    # operation.
    if operation is not False:
        try:
            oldValue = getattr(self, attrib)
        except AttributeError:
            # attribute is not set yet. Set it to None to skip
            # operation and just set value.
            oldValue = None

        # Apply operation except for the case when new or old value
        # are None or string-like
        if (value is not None and
                not isinstance(value, basestring) and
                oldValue is not None and
                not isinstance(oldValue, basestring)):
            value = numpy.array(value, float)

            # Calculate new value using operation
            if operation in ('', None):
                if (value.shape is () and
                        not isinstance(oldValue, attributeSetter)):  # scalar
                    # Preserves dimensions in case oldValue is array-like.
                    value = oldValue * 0 + value
            elif operation == '+':
                value = oldValue + value
            elif operation == '*':
                value = oldValue * value
            elif operation == '-':
                value = oldValue - value
            elif operation == '/':
                value = oldValue / value
            elif operation == '**':
                value = oldValue ** value
            elif operation == '%':
                value = oldValue % value
            else:
                msg = ('Unsupported value "%s" for operation when '
                       'setting %s in %s')
                vals = (operation, attrib, self.__class__.__name__)
                raise ValueError(msg % vals)

        elif operation not in ('', None):
            msg = ('operation %s invalid for %s (old value) and %s '
                   '(operation value)')
            vals = (operation.__repr__(), oldValue, value)
            raise TypeError(msg % vals)

    # Ok, operation or not, change the attribute in self without callback to
    # attributeSetters
    if stealth:
        self.__dict__[attrib] = value  # without logging as well
    else:
        # Trick to control logging of attributeSetter. Set logging in
        # self.autoLog
        autoLogOrig = self.autoLog  # save original value
        # set to desired logging. log=None dafaults to autoLog
        self.__dict__['autoLog'] = log or autoLogOrig and log is None
        # set attribute, calling attributeSetter if it exists
        setattr(self, attrib, value)
        # hack: if attrib was 'autoLog', do not set it back to original value!
        if attrib != 'autoLog':
            # return autoLog to original
            self.__dict__['autoLog'] = autoLogOrig


def logAttrib(obj, log, attrib, value=None):
    """Logs a change of a visual attribute on the next window.flip.
    If value=None, it will take the value of self.attrib.
    """
    # Default to autoLog if log isn't set explicitly
    if log or log is None and obj.autoLog == True:
        if value is None:
            value = getattr(obj, attrib)

        # for numpy arrays bigger than 2x2 repr is slow (up to 1ms) so just
        # say it was an array
        if isinstance(value, numpy.ndarray) \
                and (value.ndim > 2 or len(value) > 2):
            valStr = repr(type(value))
        else:
            valStr = value.__repr__()
        message = "%s: %s = %s" % (obj.name, attrib, valStr)

        try:
            obj.win.logOnFlip(message, level=logging.EXP, obj=obj)
        except AttributeError:
            # the "win" attribute only exists if sync-to-visual (e.g. stimuli)
            logging.log(message, level=logging.EXP, obj=obj)
